# Copyright (c) Meta Platforms, Inc. and affiliates.
# Copyright (c) 2025, Kylin Software, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import mmap
import os
import re

from drgn import Object
from drgn.helpers.linux.cpumask import for_each_online_cpu
from drgn.helpers.linux.mm import PFN_PHYS, TaskRss, phys_to_virt
from drgn.helpers.linux.percpu import per_cpu_ptr
from tests.linux_kernel import (
    skip_if_highmem,
    skip_unless_have_full_mm_support,
    skip_unless_have_test_disk,
    skip_unless_have_test_kmod,
)
from tests.linux_kernel.crash_commands import CrashCommandTestCase
from tests.linux_kernel.helpers.test_mm import MmTestCase
from tests.linux_kernel.helpers.test_swap import tmp_swaps


class TestBtop(CrashCommandTestCase):
    def test_single(self):
        addr = mmap.PAGESIZE * 2
        cmd = self.check_crash_command(f"btop {addr:x}")
        self.assertEqual(cmd.stdout, f"{addr:x}: 2\n")
        self.assertEqual(cmd.drgn_option.globals["phys_addr"], addr)
        self.assertEqual(cmd.drgn_option.globals["pfn"], 2)

    def test_multiple(self):
        addr1 = mmap.PAGESIZE * 2
        addr2 = mmap.PAGESIZE * 10
        cmd = self.check_crash_command(f"btop {addr1:x} {addr2:x}")
        self.assertEqual(cmd.stdout, f"{addr1:x}: 2\n{addr2:x}: a\n")
        self.assertRegex(cmd.drgn_option.stdout, r"\bfor\b.*\bin\b")
        self.assertEqual(cmd.drgn_option.globals["phys_addr"], addr2)
        self.assertEqual(cmd.drgn_option.globals["pfn"], 10)


class TestPtob(CrashCommandTestCase):
    def test_single(self):
        addr = mmap.PAGESIZE * 2
        cmd = self.check_crash_command("ptob 2")
        self.assertEqual(cmd.stdout, f"2: {addr:x}\n")
        self.assertEqual(cmd.drgn_option.globals["pfn"], 2)
        self.assertEqual(cmd.drgn_option.globals["phys_addr"], addr)

    def test_multiple(self):
        addr1 = mmap.PAGESIZE * 2
        addr2 = mmap.PAGESIZE * 10
        cmd = self.check_crash_command("ptob 2 10")
        self.assertEqual(cmd.stdout, f"2: {addr1:x}\na: {addr2:x}\n")
        self.assertRegex(cmd.drgn_option.stdout, r"\bfor\b.*\bin\b")
        self.assertEqual(cmd.drgn_option.globals["pfn"], 10)
        self.assertEqual(cmd.drgn_option.globals["phys_addr"], addr2)


class TestPtov(CrashCommandTestCase):
    @skip_unless_have_full_mm_support
    def test_phy_to_virt(self):
        """Test physical address to virtual address conversion."""
        phys_addr = 0x123
        virt_addr = phys_to_virt(self.prog, phys_addr)
        virt_addr_int = virt_addr.value_()

        cmd = self.check_crash_command(f"ptov {hex(phys_addr)}")
        self.assertRegex(cmd.stdout, r"(?m)^\s*VIRTUAL\s+PHYSICAL")
        self.assertRegex(cmd.stdout, rf"(?m)^\s*{virt_addr_int:016x}\s+{phys_addr:x}")

    def test_per_cpu_offset_single_cpu(self):
        """Test per-CPU offset conversion for a single CPU."""
        offset = 0x100
        cpu = 0
        ptr = Object(self.prog, "unsigned long", offset)
        virt_ptr = per_cpu_ptr(ptr, cpu)
        virt_int = virt_ptr.value_()

        cmd = self.check_crash_command(f"ptov {hex(offset)}:{cpu}")
        self.assertRegex(cmd.stdout, rf"(?m)^\s*PER-CPU OFFSET:\s+{offset:x}")
        self.assertRegex(cmd.stdout, r"(?m)^\s*CPU\s+VIRTUAL")
        self.assertRegex(cmd.stdout, rf"(?m)^\s*\[{cpu}\]\s+{virt_int:016x}")

    def test_per_cpu_offset_all_cpus(self):
        """Test per-CPU offset conversion for all CPUs."""
        offset = 0x200
        cmd = self.check_crash_command(f"ptov {hex(offset)}:a")

        self.assertRegex(cmd.stdout, rf"(?m)^\s*PER-CPU OFFSET:\s+{offset:x}")
        self.assertRegex(cmd.stdout, r"(?m)^\s*CPU\s+VIRTUAL")

        ptr = Object(self.prog, "unsigned long", offset)
        for cpu in for_each_online_cpu(self.prog):
            virt = per_cpu_ptr(ptr, cpu)
            self.assertRegex(cmd.stdout, rf"(?m)^\s*\[{cpu}\]\s+{virt.value_():016x}")

    def test_per_cpu_offset_cpu_list(self):
        """Test per-CPU offset conversion for a CPU list."""
        offset = 0x300
        cpus = sorted(os.sched_getaffinity(0))
        cmd = self.check_crash_command(f"ptov {hex(offset)}:{','.join(map(str, cpus))}")

        self.assertRegex(cmd.stdout, rf"(?m)^\s*PER-CPU OFFSET:\s+{offset:x}")
        self.assertRegex(cmd.stdout, r"(?m)^\s*CPU\s+VIRTUAL")

        ptr = Object(self.prog, "unsigned long", offset)
        for cpu in cpus:
            virt = per_cpu_ptr(ptr, cpu)
            self.assertRegex(cmd.stdout, rf"(?m)^\s*\[{cpu}\]\s+{virt.value_():016x}")

    def test_invalid_address(self):
        """Test invalid physical address input."""
        with self.assertRaises(Exception) as cm:
            self.check_crash_command("ptov invalid_address")
        msg = str(cm.exception).lower()
        self.assertTrue(
            "invalid literal" in msg or "base 16" in msg,
            f"Unexpected error message: {msg}",
        )

    def test_invalid_cpu_spec(self):
        """Test invalid per-CPU specifier."""
        offset = 0x400
        with self.assertRaises(Exception) as cm:
            self.check_crash_command(f"ptov {hex(offset)}:invalid")
        msg = str(cm.exception).lower()
        self.assertIn("invalid cpuspec", msg, f"Unexpected error message: {msg}")


@skip_unless_have_full_mm_support
class TestVtop(CrashCommandTestCase, MmTestCase):
    @skip_unless_have_test_kmod
    def test_kernel(self):
        virt_addr = self.prog["drgn_test_va"].value_()
        phys_addr = self.prog["drgn_test_pa"].value_()

        cmd = self.check_crash_command(f"vtop -k {virt_addr:x}")
        self.assertIn(f"{phys_addr:x}", cmd.stdout)

        self.assertIn("follow_phys", cmd.drgn_option.stdout)
        self.assertEqual(cmd.drgn_option.globals["virtual_address"], virt_addr)
        self.assertEqual(cmd.drgn_option.globals["physical_address"], phys_addr)

    @skip_unless_have_test_kmod
    def test_kernel_multiple(self):
        virt_addrs = [
            self.prog["drgn_test_va"].value_(),
            self.prog["drgn_test_vmalloc_va"].value_(),
        ]
        phys_addrs = [
            self.prog["drgn_test_pa"].value_(),
            self.prog["drgn_test_vmalloc_pa"].value_(),
        ]

        cmd = self.check_crash_command(f"vtop -k {virt_addrs[0]:x} {virt_addrs[1]:x}")
        for phys_addr in phys_addrs:
            self.assertIn(f"{phys_addr:x}", cmd.stdout)

        self.assertIn("follow_phys", cmd.drgn_option.stdout)
        for virt_addr in virt_addrs:
            self.assertIn(hex(virt_addr), cmd.drgn_option.stdout)
        self.assertIn(cmd.drgn_option.globals["virtual_address"], virt_addrs)
        self.assertIn(cmd.drgn_option.globals["physical_address"], phys_addrs)

    @skip_if_highmem
    def test_user(self):
        with self.map_pages() as (_, address, pfns):
            phys_addr = PFN_PHYS(self.prog, pfns[0]).value_()
            cmd = self.check_crash_command(f"vtop -u -c {os.getpid()} {address:x}")
        self.assertIn(f"{phys_addr:x}", cmd.stdout)

        self.assertIn("follow_phys", cmd.drgn_option.stdout)
        self.assertEqual(cmd.drgn_option.globals["virtual_address"], address)
        self.assertEqual(cmd.drgn_option.globals["physical_address"], phys_addr)

    @skip_if_highmem
    def test_user_multiple(self):
        with self.map_pages() as (_, address, pfns):
            virt_addrs = [address, address + mmap.PAGESIZE]
            phys_addrs = [PFN_PHYS(self.prog, pfns[i]).value_() for i in range(2)]
            cmd = self.check_crash_command(
                f"vtop -u -c {os.getpid()} {virt_addrs[0]:x} {virt_addrs[1]:x}"
            )
        for phys_addr in phys_addrs:
            self.assertIn(f"{phys_addr:x}", cmd.stdout)

        self.assertIn("follow_phys", cmd.drgn_option.stdout)
        for virt_addr in virt_addrs:
            self.assertIn(hex(virt_addr), cmd.drgn_option.stdout)
        self.assertIn(cmd.drgn_option.globals["virtual_address"], virt_addrs)
        self.assertIn(cmd.drgn_option.globals["physical_address"], phys_addrs)

    @skip_unless_have_test_kmod
    def test_guess(self):
        virt_addr = self.prog["drgn_test_va"].value_()
        phys_addr = self.prog["drgn_test_pa"].value_()

        cmd = self.check_crash_command(f"vtop {virt_addr:x}")
        self.assertTrue(
            f"{phys_addr:x}" in cmd.stdout or "ambiguous address" in cmd.stderr
        )

        self.assertIn("follow_phys(init_mm", cmd.drgn_option.stdout)
        self.assertIn("follow_phys(mm", cmd.drgn_option.stdout)
        self.assertEqual(cmd.drgn_option.globals["virtual_address"], virt_addr)


class TestSwap(CrashCommandTestCase):
    @skip_unless_have_test_disk
    def test_swap(self):
        with tmp_swaps() as swaps:
            cmd = self.check_crash_command("swap")
            for path, is_file in swaps:
                type = "FILE" if is_file else "PARTITION"
                self.assertRegex(cmd.stdout, f"{type} .* {re.escape(str(path))}")
        self.assertEqual(
            cmd.drgn_option.globals["si"].type_.type_name(), "struct swap_info_struct *"
        )
        self.assertIn("pages", cmd.drgn_option.globals)
        self.assertIn("used_pages", cmd.drgn_option.globals)
        self.assertIn("priority", cmd.drgn_option.globals)
        self.assertIsInstance(cmd.drgn_option.globals["path"], bytes)


class TestVm(CrashCommandTestCase):
    def _test_common(self, cmd):
        self.assertRegex(
            cmd.stdout,
            r"\bMM\s+PGD\s+RSS\s+TOTAL_VM(?:\s+[0-9a-f]+){2}(?:\s+[0-9]+k){2}\b",
        )
        self.assertRegex(
            cmd.stdout, r"\bVMA\s+START\s+END\s+FLAGS\s+FILE(?:\s+[0-9a-f]+){4}"
        )
        self.assertRegex(cmd.stdout, r"\bVMA(?s:.)*/")  # Check for any file path.

    def _test_drgn_option_common(self, cmd):
        for variable in (
            "pid",
            "task",
            "command",
            "mm",
            "pgd",
            "vma",
            "start",
            "end",
            "flags",
        ):
            with self.subTest(variable=variable):
                self.assertIsInstance(cmd.drgn_option.globals[variable], Object)
        self.assertIsInstance(cmd.drgn_option.globals["cpu"], int)
        self.assertIsInstance(cmd.drgn_option.globals["rss"], TaskRss)
        self.assertIsInstance(cmd.drgn_option.globals["file"], bytes)

    def test_no_args(self):
        self.run_crash_command("set -p")

        cmd = self.check_crash_command("vm")
        self.assertIn(f"PID: {os.getpid()}", cmd.stdout)
        self._test_common(cmd)

        self._test_drgn_option_common(cmd)

    def test_tasks(self):
        cmd = self.check_crash_command(f"vm 1 {os.getpid()}")
        foreach_cmd = self.check_crash_command(
            f"foreach 1 {os.getpid()} vm", mode="capture"
        )

        for c in (cmd, foreach_cmd):
            self.assertIn("PID: 1", c.stdout)
            self.assertIn(f"PID: {os.getpid()}", c.stdout)
            self._test_common(c)

        self._test_drgn_option_common(cmd)

        self.assertEqual(cmd.drgn_option.stdout, foreach_cmd.drgn_option.stdout)

    def test_kernel_thread(self):
        cmd = self.check_crash_command("vm 2")

        self.assertRegex(cmd.stdout, r"\bMM\s+PGD\s+RSS\s+TOTAL_VM\s+0\s+0\s+0k\s+0k\b")
        self.assertNotIn("VMA", cmd.stdout)

        self.assertFalse(cmd.drgn_option.globals["mm"])



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

from drgn.helpers.linux import (
    plist_first_entry,
    plist_for_each,
    plist_for_each_entry,
    plist_head_empty,
    plist_last_entry,
    plist_node_empty,
)
from tests.linux_kernel import LinuxKernelTestCase, skip_unless_have_test_kmod


@skip_unless_have_test_kmod
class TestPlist(LinuxKernelTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.empty = cls.prog["drgn_test_empty_plist"].address_of_()
        cls.full = cls.prog["drgn_test_full_plist"].address_of_()
        cls.entries = cls.prog["drgn_test_plist_entries"]
        cls.num_entries = 3

    def node(self, n):
        return self.entries[n].node.address_of_()

    def entry(self, n):
        return self.entries[n].address_of_()

    def test_plist_head_empty(self):
        self.assertTrue(plist_head_empty(self.empty))
        self.assertFalse(plist_head_empty(self.full))

    def test_plist_node_empty(self):
        self.assertTrue(
            plist_node_empty(self.prog["drgn_test_empty_plist_node"].address_of_())
        )
        self.assertFalse(plist_node_empty(self.node(0)))

    def test_plist_first_entry(self):
        self.assertEqual(
            plist_first_entry(self.full, "struct drgn_test_plist_entry", "node"),
            self.entry(0),
        )

    def test_plist_last_entry(self):
        self.assertEqual(
            plist_last_entry(self.full, "struct drgn_test_plist_entry", "node"),
            self.entry(self.num_entries - 1),
        )

    def test_plist_for_each(self):
        self.assertEqual(list(plist_for_each(self.empty)), [])
        self.assertEqual(
            list(plist_for_each(self.full)),
            [self.node(i) for i in range(self.num_entries)],
        )

    def test_plist_for_each_entry(self):
        self.assertEqual(
            list(
                plist_for_each_entry("struct drgn_test_plist_entry", self.empty, "node")
            ),
            [],
        )
        self.assertEqual(
            list(
                plist_for_each_entry("struct drgn_test_plist_entry", self.full, "node")
            ),
            [self.entry(i) for i in range(self.num_entries)],
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import array
import ctypes
import os
import re
import sys
from typing import NamedTuple, Tuple

from _drgn_util.platform import SYS
from tests.linux_kernel import _check_ctypes_syscall, _syscall

# enum bpf_cmd
BPF_MAP_CREATE = 0
BPF_MAP_LOOKUP_ELEM = 1
BPF_MAP_UPDATE_ELEM = 2
BPF_MAP_DELETE_ELEM = 3
BPF_MAP_GET_NEXT_KEY = 4
BPF_PROG_LOAD = 5
BPF_OBJ_PIN = 6
BPF_OBJ_GET = 7
BPF_PROG_ATTACH = 8
BPF_PROG_DETACH = 9
BPF_PROG_TEST_RUN = 10
BPF_PROG_RUN = BPF_PROG_TEST_RUN
BPF_PROG_GET_NEXT_ID = 11
BPF_MAP_GET_NEXT_ID = 12
BPF_PROG_GET_FD_BY_ID = 13
BPF_MAP_GET_FD_BY_ID = 14
BPF_OBJ_GET_INFO_BY_FD = 15
BPF_PROG_QUERY = 16
BPF_RAW_TRACEPOINT_OPEN = 17
BPF_BTF_LOAD = 18
BPF_BTF_GET_FD_BY_ID = 19
BPF_TASK_FD_QUERY = 20
BPF_MAP_LOOKUP_AND_DELETE_ELEM = 21
BPF_MAP_FREEZE = 22
BPF_BTF_GET_NEXT_ID = 23
BPF_MAP_LOOKUP_BATCH = 24
BPF_MAP_LOOKUP_AND_DELETE_BATCH = 25
BPF_MAP_UPDATE_BATCH = 26
BPF_MAP_DELETE_BATCH = 27
BPF_LINK_CREATE = 28
BPF_LINK_UPDATE = 29
BPF_LINK_GET_FD_BY_ID = 30
BPF_LINK_GET_NEXT_ID = 31
BPF_ENABLE_STATS = 32
BPF_ITER_CREATE = 33
BPF_LINK_DETACH = 34
BPF_PROG_BIND_MAP = 35

# enum bpf_map_type
BPF_MAP_TYPE_UNSPEC = 0
BPF_MAP_TYPE_HASH = 1
BPF_MAP_TYPE_ARRAY = 2
BPF_MAP_TYPE_PROG_ARRAY = 3
BPF_MAP_TYPE_PERF_EVENT_ARRAY = 4
BPF_MAP_TYPE_PERCPU_HASH = 5
BPF_MAP_TYPE_PERCPU_ARRAY = 6
BPF_MAP_TYPE_STACK_TRACE = 7
BPF_MAP_TYPE_CGROUP_ARRAY = 8
BPF_MAP_TYPE_LRU_HASH = 9
BPF_MAP_TYPE_LRU_PERCPU_HASH = 10
BPF_MAP_TYPE_LPM_TRIE = 11
BPF_MAP_TYPE_ARRAY_OF_MAPS = 12
BPF_MAP_TYPE_HASH_OF_MAPS = 13
BPF_MAP_TYPE_DEVMAP = 14
BPF_MAP_TYPE_SOCKMAP = 15
BPF_MAP_TYPE_CPUMAP = 16
BPF_MAP_TYPE_XSKMAP = 17
BPF_MAP_TYPE_SOCKHASH = 18
BPF_MAP_TYPE_CGROUP_STORAGE = 19
BPF_MAP_TYPE_REUSEPORT_SOCKARRAY = 20
BPF_MAP_TYPE_PERCPU_CGROUP_STORAGE = 21
BPF_MAP_TYPE_QUEUE = 22
BPF_MAP_TYPE_STACK = 23
BPF_MAP_TYPE_SK_STORAGE = 24
BPF_MAP_TYPE_DEVMAP_HASH = 25
BPF_MAP_TYPE_STRUCT_OPS = 26
BPF_MAP_TYPE_RINGBUF = 27
BPF_MAP_TYPE_INODE_STORAGE = 28
BPF_MAP_TYPE_TASK_STORAGE = 29
BPF_MAP_TYPE_BLOOM_FILTER = 30

# enum bpf_prog_type
BPF_PROG_TYPE_UNSPEC = 0
BPF_PROG_TYPE_SOCKET_FILTER = 1
BPF_PROG_TYPE_KPROBE = 2
BPF_PROG_TYPE_SCHED_CLS = 3
BPF_PROG_TYPE_SCHED_ACT = 4
BPF_PROG_TYPE_TRACEPOINT = 5
BPF_PROG_TYPE_XDP = 6
BPF_PROG_TYPE_PERF_EVENT = 7
BPF_PROG_TYPE_CGROUP_SKB = 8
BPF_PROG_TYPE_CGROUP_SOCK = 9
BPF_PROG_TYPE_LWT_IN = 10
BPF_PROG_TYPE_LWT_OUT = 11
BPF_PROG_TYPE_LWT_XMIT = 12
BPF_PROG_TYPE_SOCK_OPS = 13
BPF_PROG_TYPE_SK_SKB = 14
BPF_PROG_TYPE_CGROUP_DEVICE = 15
BPF_PROG_TYPE_SK_MSG = 16
BPF_PROG_TYPE_RAW_TRACEPOINT = 17
BPF_PROG_TYPE_CGROUP_SOCK_ADDR = 18
BPF_PROG_TYPE_LWT_SEG6LOCAL = 19
BPF_PROG_TYPE_LIRC_MODE2 = 20
BPF_PROG_TYPE_SK_REUSEPORT = 21
BPF_PROG_TYPE_FLOW_DISSECTOR = 22
BPF_PROG_TYPE_CGROUP_SYSCTL = 23
BPF_PROG_TYPE_RAW_TRACEPOINT_WRITABLE = 24
BPF_PROG_TYPE_CGROUP_SOCKOPT = 25
BPF_PROG_TYPE_TRACING = 26
BPF_PROG_TYPE_STRUCT_OPS = 27
BPF_PROG_TYPE_EXT = 28
BPF_PROG_TYPE_LSM = 29
BPF_PROG_TYPE_SK_LOOKUP = 30
BPF_PROG_TYPE_SYSCALL = 31

# enum bpf_attach_type
BPF_CGROUP_INET_INGRESS = 0
BPF_CGROUP_INET_EGRESS = 1
BPF_CGROUP_INET_SOCK_CREATE = 2
BPF_CGROUP_SOCK_OPS = 3
BPF_SK_SKB_STREAM_PARSER = 4
BPF_SK_SKB_STREAM_VERDICT = 5
BPF_CGROUP_DEVICE = 6
BPF_SK_MSG_VERDICT = 7
BPF_CGROUP_INET4_BIND = 8
BPF_CGROUP_INET6_BIND = 9
BPF_CGROUP_INET4_CONNECT = 10
BPF_CGROUP_INET6_CONNECT = 11
BPF_CGROUP_INET4_POST_BIND = 12
BPF_CGROUP_INET6_POST_BIND = 13
BPF_CGROUP_UDP4_SENDMSG = 14
BPF_CGROUP_UDP6_SENDMSG = 15
BPF_LIRC_MODE2 = 16
BPF_FLOW_DISSECTOR = 17
BPF_CGROUP_SYSCTL = 18
BPF_CGROUP_UDP4_RECVMSG = 19
BPF_CGROUP_UDP6_RECVMSG = 20
BPF_CGROUP_GETSOCKOPT = 21
BPF_CGROUP_SETSOCKOPT = 22
BPF_TRACE_RAW_TP = 23
BPF_TRACE_FENTRY = 24
BPF_TRACE_FEXIT = 25
BPF_MODIFY_RETURN = 26
BPF_LSM_MAC = 27
BPF_TRACE_ITER = 28
BPF_CGROUP_INET4_GETPEERNAME = 29
BPF_CGROUP_INET6_GETPEERNAME = 30
BPF_CGROUP_INET4_GETSOCKNAME = 31
BPF_CGROUP_INET6_GETSOCKNAME = 32
BPF_XDP_DEVMAP = 33
BPF_CGROUP_INET_SOCK_RELEASE = 34
BPF_XDP_CPUMAP = 35
BPF_SK_LOOKUP = 36
BPF_XDP = 37
BPF_SK_SKB_VERDICT = 38
BPF_SK_REUSEPORT_SELECT = 39
BPF_SK_REUSEPORT_SELECT_OR_MIGRATE = 40
BPF_PERF_EVENT = 41

# Flags for BPF_PROG_ATTACH.
BPF_F_ALLOW_OVERRIDE = 1 << 0
BPF_F_ALLOW_MULTI = 1 << 1
BPF_F_REPLACE = 1 << 2


# Instruction classes.
BPF_LD = 0x00
BPF_LDX = 0x01
BPF_ST = 0x02
BPF_STX = 0x03
BPF_ALU = 0x04
BPF_JMP = 0x05
BPF_RET = 0x06
BPF_JMP32 = 0x06
BPF_MISC = 0x07
BPF_ALU64 = 0x07

# ld/ldx fields.
BPF_W = 0x00
BPF_H = 0x08
BPF_B = 0x10
BPF_DW = 0x18
BPF_IMM = 0x00
BPF_ABS = 0x20
BPF_IND = 0x40
BPF_MEM = 0x60
BPF_LEN = 0x80
BPF_MSH = 0xA0
BPF_MEMSX = 0x80
BPF_ATOMIC = 0xC0
BPF_XADD = 0xC0

# alu/jmp fields.
BPF_ADD = 0x00
BPF_SUB = 0x10
BPF_MUL = 0x20
BPF_DIV = 0x30
BPF_OR = 0x40
BPF_AND = 0x50
BPF_LSH = 0x60
BPF_RSH = 0x70
BPF_NEG = 0x80
BPF_MOD = 0x90
BPF_XOR = 0xA0

BPF_JA = 0x00
BPF_JEQ = 0x10
BPF_JGT = 0x20
BPF_JGE = 0x30
BPF_JSET = 0x40
BPF_K = 0x00
BPF_X = 0x08

BPF_MOV = 0xB0
BPF_ARSH = 0xC0

# Change endianness of a register.
BPF_END = 0xD0
BPF_TO_LE = 0x00
BPF_TO_BE = 0x08
BPF_FROM_LE = BPF_TO_LE
BPF_FROM_BE = BPF_TO_BE

# jmp encodings.
BPF_JNE = 0x50
BPF_JLT = 0xA0
BPF_JLE = 0xB0
BPF_JSGT = 0x60
BPF_JSGE = 0x70
BPF_JSLT = 0xC0
BPF_JSLE = 0xD0
BPF_JCOND = 0xE0
BPF_CALL = 0x80
BPF_EXIT = 0x90

# atomic op type fields.
BPF_FETCH = 0x01
BPF_XCHG = 0xE0 | BPF_FETCH
BPF_CMPXCHG = 0xF0 | BPF_FETCH

BPF_LOAD_ACQ = 0x100
BPF_STORE_REL = 0x110

# Register numbers.
BPF_REG_0 = 0
BPF_REG_1 = 1
BPF_REG_2 = 2
BPF_REG_3 = 3
BPF_REG_4 = 4
BPF_REG_5 = 5
BPF_REG_6 = 6
BPF_REG_7 = 7
BPF_REG_8 = 8
BPF_REG_9 = 9
BPF_REG_10 = 10

MAX_BPF_REG = 11


# Instruction encoding.
def bpf_insn(*, code: int, dst_reg: int, src_reg: int, off: int, imm: int) -> int:
    if code < 0 or code > 255:
        raise ValueError(f"code {code} is out of range")
    if dst_reg < 0 or dst_reg > 15:
        raise ValueError(f"dst_reg {dst_reg} is out of range")
    if src_reg < 0 or src_reg > 15:
        raise ValueError(f"src_reg {src_reg} is out of range")
    if off < -32768 or off > 32767:
        raise ValueError(f"off {off} is out of range")
    # Allow signed and unsigned 32-bit values.
    if imm < -(2**31) or imm >= 2**32:
        raise ValueError(f"imm {imm} is out of range")
    if sys.byteorder == "little":
        return (
            code
            | (dst_reg << 8)
            | (src_reg << 12)
            | ((off & 0xFFFF) << 16)
            | ((imm & 0xFFFFFFFF) << 32)
        )
    else:
        return (
            (code << 56)
            | (dst_reg << 52)
            | (src_reg << 48)
            | ((off & 0xFFFF) << 32)
            | (imm & 0xFFFFFFFF)
        )


# Instructions.
def BPF_MOV64_IMM(dst: int, imm: int) -> int:
    return bpf_insn(
        code=BPF_ALU64 | BPF_MOV | BPF_K,
        dst_reg=dst,
        src_reg=0,
        off=0,
        imm=imm,
    )


BPF_PSEUDO_MAP_FD = 1
BPF_PSEUDO_MAP_IDX = 5
BPF_PSEUDO_MAP_VALUE = 2
BPF_PSEUDO_MAP_IDX_VALUE = 6
BPF_PSEUDO_BTF_ID = 3
BPF_PSEUDO_FUNC = 4


def BPF_LD_IMM64_RAW(dst: int, src: int, imm: int) -> Tuple[int, int]:
    # Allow signed and unsigned 64-bit values.
    if imm < -(2**63) or imm >= 2**64:
        raise ValueError(f"imm {imm} is out of range")
    return (
        bpf_insn(
            code=BPF_LD | BPF_DW | BPF_IMM,
            dst_reg=dst,
            src_reg=src,
            off=0,
            imm=imm & 0xFFFFFFFF,
        ),
        bpf_insn(code=0, dst_reg=0, src_reg=0, off=0, imm=(imm >> 32) & 0xFFFFFFFF),
    )


def BPF_LD_MAP_FD(dst: int, map_fd: int) -> int:
    return BPF_LD_IMM64_RAW(dst, BPF_PSEUDO_MAP_FD, map_fd)


def BPF_EXIT_INSN() -> int:
    return bpf_insn(
        code=BPF_JMP | BPF_EXIT,
        dst_reg=0,
        src_reg=0,
        off=0,
        imm=0,
    )


class _bpf_attr_map_create(ctypes.Structure):
    _fields_ = (
        ("map_type", ctypes.c_uint32),
        ("key_size", ctypes.c_uint32),
        ("value_size", ctypes.c_uint32),
        ("max_entries", ctypes.c_uint32),
    )


_BPF_OBJ_NAME_LEN = 16


class _bpf_attr_prog_load(ctypes.Structure):
    _fields_ = (
        ("prog_type", ctypes.c_uint32),
        ("insn_cnt", ctypes.c_uint32),
        ("insns", ctypes.c_uint64),
        ("license", ctypes.c_uint64),
        ("log_level", ctypes.c_uint32),
        ("log_size", ctypes.c_uint32),
        ("log_buf", ctypes.c_uint64),
        ("kern_version", ctypes.c_uint32),
        ("prog_flags", ctypes.c_uint32),
        ("prog_name", ctypes.c_char * _BPF_OBJ_NAME_LEN),
        ("prog_ifindex", ctypes.c_uint32),
        ("expected_attach_type", ctypes.c_uint32),
    )


class _bpf_attr_attach(ctypes.Structure):
    _fields_ = (
        ("target_fd", ctypes.c_uint32),
        ("attach_bpf_fd", ctypes.c_uint32),
        ("attach_type", ctypes.c_uint32),
        ("attach_flags", ctypes.c_uint32),
        ("replace_bpf_fd", ctypes.c_uint32),
    )


class _bpf_attr_get_id(ctypes.Structure):
    _fields_ = (
        ("id", ctypes.c_uint32),
        ("next_id", ctypes.c_uint32),
        ("open_flags", ctypes.c_uint32),
    )


class _bpf_attr_obj_get_info_by_fd(ctypes.Structure):
    _fields_ = (
        ("bpf_fd", ctypes.c_uint32),
        ("info_len", ctypes.c_uint32),
        ("info", ctypes.c_uint64),
    )


class _bpf_attr_link_create(ctypes.Structure):
    _fields_ = (
        ("prog_fd", ctypes.c_uint32),
        ("target_fd", ctypes.c_uint32),
        ("attach_type", ctypes.c_uint32),
        ("flags", ctypes.c_uint32),
    )


class _bpf_attr(ctypes.Union):
    _fields_ = (
        ("_map_create", _bpf_attr_map_create),
        ("_prog_load", _bpf_attr_prog_load),
        ("_attach", _bpf_attr_attach),
        ("_get_id", _bpf_attr_get_id),
        ("_obj_get_info_by_fd", _bpf_attr_obj_get_info_by_fd),
        ("link_create", _bpf_attr_link_create),
    )
    _anonymous_ = (
        "_map_create",
        "_prog_load",
        "_attach",
        "_get_id",
        "_obj_get_info_by_fd",
    )


try:
    _SYS_bpf = ctypes.c_long(SYS["bpf"])
except KeyError:
    _SYS_bpf = None
_sizeof_bpf_attr = ctypes.c_uint(ctypes.sizeof(_bpf_attr))


def _bpf(cmd, attr):
    return _check_ctypes_syscall(
        _syscall(_SYS_bpf, ctypes.c_int(cmd), ctypes.byref(attr), _sizeof_bpf_attr)
    )


def bpf_map_create(map_type, key_size, value_size, max_entries):
    attr = _bpf_attr()
    attr.map_type = map_type
    attr.key_size = key_size
    attr.value_size = value_size
    attr.max_entries = max_entries
    return _bpf(BPF_MAP_CREATE, attr)


_LOG_BUF_SIZE = 65536


class BpfVerifierError(OSError):
    def __init__(self, *args, log) -> None:
        super().__init__(*args)
        self.log = log

    def __str__(self) -> str:
        return f"{super().__str__()}\n{self.log}"


def bpf_prog_load(prog_type, insns, license, expected_attach_type=0):
    attr = _bpf_attr()
    attr.prog_type = prog_type
    insns_array = array.array("Q", insns)
    attr.insns, attr.insn_cnt = insns_array.buffer_info()
    ctypes_license = ctypes.c_char_p(license)
    attr.license = ctypes.cast(ctypes_license, ctypes.c_void_p).value

    attr.log_level = 1
    attr.log_size = _LOG_BUF_SIZE
    log_buf = ctypes.create_string_buffer(_LOG_BUF_SIZE)
    attr.log_buf = ctypes.addressof(log_buf)

    if prog_type == BPF_PROG_TYPE_KPROBE:
        # Before Linux kernel commit 6c4fc209fcf9 ("bpf: remove useless version
        # check for prog load") (in v5.0), for BPF_PROG_TYPE_KPROBE,
        # kern_version must match the kernel's LINUX_VERSION_CODE.
        match = re.match(r"([0-9]+)\.([0-9]+)\.([0-9]+)", os.uname().release)
        if match:
            attr.kern_version = (
                (int(match.group(1)) << 16)
                | (int(match.group(2)) << 8)
                | min(int(match.group(3)), 255)
            )

    attr.expected_attach_type = expected_attach_type

    ret = _syscall(
        _SYS_bpf, ctypes.c_int(BPF_PROG_LOAD), ctypes.byref(attr), _sizeof_bpf_attr
    )
    if ret == -1:
        errno = ctypes.get_errno()
        log = log_buf.value
        if log:
            raise BpfVerifierError(errno, os.strerror(errno), log=log.decode())
        else:
            raise OSError(errno, os.strerror(errno))
    return ret


def bpf_prog_attach(target_fd, attach_bpf_fd, attach_type, attach_flags=0):
    attr = _bpf_attr()
    attr.target_fd = target_fd
    attr.attach_bpf_fd = attach_bpf_fd
    attr.attach_type = attach_type
    attr.attach_flags = attach_flags
    _bpf(BPF_PROG_ATTACH, attr)


def _bpf_get_ids(cmd):
    attr = _bpf_attr()
    attr.id = 0
    while True:
        try:
            _bpf(cmd, attr)
        except FileNotFoundError:
            break
        yield attr.next_id
        attr.id = attr.next_id


def bpf_map_ids():
    return _bpf_get_ids(BPF_MAP_GET_NEXT_ID)


def bpf_prog_ids():
    return _bpf_get_ids(BPF_PROG_GET_NEXT_ID)


def bpf_btf_ids():
    return _bpf_get_ids(BPF_BTF_GET_NEXT_ID)


def bpf_link_ids():
    return _bpf_get_ids(BPF_LINK_GET_NEXT_ID)


_BPF_TAG_SIZE = 8


class _bpf_prog_info(ctypes.Structure):
    _fields_ = (
        ("type", ctypes.c_uint32),
        ("id", ctypes.c_uint32),
        ("tag", ctypes.c_uint8 * _BPF_TAG_SIZE),
        ("jited_prog_len", ctypes.c_uint32),
        ("xlated_prog_len", ctypes.c_uint32),
        ("jited_prog_insns", ctypes.c_uint64),
        ("xlated_prog_insns", ctypes.c_uint64),
    )


class BpfProgInfo(NamedTuple):
    type: int
    id: int
    tag: bytes


def bpf_prog_get_info_by_fd(bpf_fd):
    attr = _bpf_attr()
    attr.bpf_fd = bpf_fd
    info = _bpf_prog_info()
    attr.info_len = ctypes.sizeof(info)
    attr.info = ctypes.addressof(info)
    _bpf(BPF_OBJ_GET_INFO_BY_FD, attr)
    return BpfProgInfo(type=info.type, id=info.id, tag=bytes(info.tag))


class _bpf_map_info(ctypes.Structure):
    _fields_ = (
        ("type", ctypes.c_uint32),
        ("id", ctypes.c_uint32),
        ("key_size", ctypes.c_uint32),
        ("value_size", ctypes.c_uint32),
        ("max_entries", ctypes.c_uint32),
        ("map_flags", ctypes.c_uint32),
        ("name", ctypes.c_char * _BPF_OBJ_NAME_LEN),
        ("ifindex", ctypes.c_uint32),
        ("btf_vmlinux_value_type_id", ctypes.c_uint32),
        ("netns_dev", ctypes.c_uint64),
        ("netns_ino", ctypes.c_uint64),
        ("btf_id", ctypes.c_uint32),
        ("btf_key_type_id", ctypes.c_uint32),
        ("btf_value_type_id", ctypes.c_uint32),
        ("btf_vmlinux_id", ctypes.c_uint32),
        ("map_extra", ctypes.c_uint64),
        ("hash", ctypes.c_uint64),
        ("hash_size", ctypes.c_uint32),
    )


class BpfMapInfo(NamedTuple):
    type: int
    id: int
    key_size: int
    value_size: int
    max_entries: int
    map_flags: int
    name: bytes
    ifindex: int
    btf_vmlinux_value_type_id: int
    netns_dev: int
    netns_ino: int
    btf_id: int
    btf_key_type_id: int
    btf_value_type_id: int
    btf_vmlinux_id: int
    map_extra: int
    hash: int
    hash_size: int


def bpf_map_get_info_by_fd(bpf_fd):
    attr = _bpf_attr()
    attr.bpf_fd = bpf_fd
    info = _bpf_map_info()
    attr.info_len = ctypes.sizeof(info)
    attr.info = ctypes.addressof(info)
    _bpf(BPF_OBJ_GET_INFO_BY_FD, attr)
    return BpfMapInfo(
        type=info.type,
        id=info.id,
        key_size=info.key_size,
        value_size=info.value_size,
        max_entries=info.max_entries,
        map_flags=info.map_flags,
        name=info.name,
        ifindex=info.ifindex,
        btf_vmlinux_value_type_id=info.btf_vmlinux_value_type_id,
        netns_dev=info.netns_dev,
        netns_ino=info.netns_ino,
        btf_id=info.btf_id,
        btf_key_type_id=info.btf_key_type_id,
        btf_value_type_id=info.btf_value_type_id,
        btf_vmlinux_id=info.btf_vmlinux_id,
        map_extra=info.map_extra,
        hash=info.hash,
        hash_size=info.hash_size,
    )


def bpf_link_create(prog_fd, target_fd, attach_type, flags=0):
    attr = _bpf_attr()
    attr.link_create.prog_fd = prog_fd
    attr.link_create.target_fd = target_fd
    attr.link_create.attach_type = attach_type
    attr.link_create.flags = flags
    return _bpf(BPF_LINK_CREATE, attr)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import os
import re
import shutil
import unittest.mock

from drgn import Object
from drgn.commands import CommandArgumentError, CommandError
from drgn.commands._crash.common import crash_get_context
from drgn.helpers.linux.pid import find_task
from tests.linux_kernel.crash_commands import CrashCommandTestCase


class TestSet(CrashCommandTestCase):
    def test_no_options(self):
        cmd = self.check_crash_command("set")
        self.assertIn("CPU:", cmd.stdout)
        for name in "task", "pid", "comm", "thread_info", "cpu", "state":
            self.assertIn(name, cmd.drgn_option.globals)

    def test_pid(self):
        cmd = self.check_crash_command("set 1")
        self.assertIn("PID: 1", cmd.stdout)
        self.assertEqual(crash_get_context(self.prog).pid.value_(), 1)
        self.assertEqual(cmd.drgn_option.globals["task"].pid, 1)

    def test_task(self):
        cmd = self.check_crash_command(f"set {hex(self.prog['init_task'].address_)}")
        self.assertIn("PID: 0", cmd.stdout)
        self.assertEqual(
            crash_get_context(self.prog), self.prog["init_task"].address_of_()
        )
        self.assertEqual(
            cmd.drgn_option.globals["task"], self.prog["init_task"].address_of_()
        )

    def test_cpu(self):
        cpu = os.cpu_count() - 1
        old_affinity = os.sched_getaffinity(0)
        os.sched_setaffinity(0, (cpu,))
        try:
            cmd = self.check_crash_command(f"set -c {cpu}")
        finally:
            os.sched_setaffinity(0, old_affinity)
        self.assertIn(f"PID: {os.getpid()}", cmd.stdout)
        task = find_task(self.prog, os.getpid())
        self.assertEqual(crash_get_context(self.prog), task)
        self.assertEqual(cmd.drgn_option.globals["task"], task)

    def test_panic(self):
        cmd = self.check_crash_command("set -p")
        task = find_task(self.prog, os.getpid())
        self.assertIn(f"PID: {os.getpid()}", cmd.stdout)
        self.assertEqual(
            crash_get_context(self.prog), find_task(self.prog, os.getpid())
        )
        self.assertEqual(cmd.drgn_option.globals["task"], task)

    def test_foreach(self):
        self.run_crash_command(f"set {hex(self.prog['init_task'].address_)}")
        cmd = self.check_crash_command(f"foreach 1 {os.getpid()} set")
        # Task information gets printed twice (once in the header and once in
        # the sys format).
        self.assertEqual(len(re.findall(r"\bPID: 1\b", cmd.stdout)), 2)
        self.assertEqual(len(re.findall(rf"\bPID: {os.getpid()}\b", cmd.stdout)), 2)
        # foreach set shouldn't change the context.
        self.assertEqual(
            crash_get_context(self.prog), self.prog["init_task"].address_of_()
        )

        for variable in ("task", "pid", "comm", "thread_info"):
            with self.subTest(variable=variable):
                self.assertIsInstance(cmd.drgn_option.globals[variable], Object)
        self.assertIsInstance(cmd.drgn_option.globals["cpu"], int)
        self.assertIsInstance(cmd.drgn_option.globals["state"], str)

    def test_scroll_on_off(self):
        try:
            old_crash_scroll = self.prog.config["crash_scroll"]
        except KeyError:
            self.addCleanup(self.prog.config.pop, "crash_scroll", None)
        else:
            self.addCleanup(
                self.prog.config.__setitem__, "crash_scroll", old_crash_scroll
            )

        cmd = self.run_crash_command("set scroll on")
        self.assertIn("scroll: on", cmd.stdout)
        cmd = self.run_crash_command("set scroll")
        self.assertIn("scroll: on", cmd.stdout)

        cmd = self.run_crash_command("set scroll off")
        self.assertIn("scroll: off", cmd.stdout)
        cmd = self.run_crash_command("set scroll")
        self.assertIn("scroll: off", cmd.stdout)

    def test_scroll_less_more(self):
        try:
            old_crash_pager = self.prog.config["crash_pager"]
        except KeyError:
            self.addCleanup(self.prog.config.pop, "crash_pager", None)
        else:
            self.addCleanup(
                self.prog.config.__setitem__, "crash_pager", old_crash_pager
            )

        def only_less(cmd):
            if cmd == "less":
                return "/usr/bin/less"
            elif cmd == "more":
                return None
            else:
                return unittest.mock.DEFAULT

        with unittest.mock.patch(
            "shutil.which", side_effect=only_less, wraps=shutil.which
        ):
            self.prog.config.pop("crash_pager", None)
            cmd = self.run_crash_command("set scroll")
            self.assertIn("/usr/bin/less", cmd.stdout)

            cmd = self.run_crash_command("set scroll less")
            self.assertIn("/usr/bin/less", cmd.stdout)
            cmd = self.run_crash_command("set scroll")
            self.assertIn("/usr/bin/less", cmd.stdout)

            with self.assertRaisesRegex(CommandError, "pager not found"):
                self.run_crash_command("set scroll more")

        def only_more(cmd):
            if cmd == "less":
                return None
            elif cmd == "more":
                return "/bin/more"
            else:
                return unittest.mock.DEFAULT

        with unittest.mock.patch(
            "shutil.which", side_effect=only_more, wraps=shutil.which
        ):
            self.prog.config.pop("crash_pager", None)
            cmd = self.run_crash_command("set scroll")
            self.assertIn("/bin/more", cmd.stdout)

            cmd = self.run_crash_command("set scroll more")
            self.assertIn("/bin/more", cmd.stdout)
            cmd = self.run_crash_command("set scroll")
            self.assertIn("/bin/more", cmd.stdout)

            with self.assertRaisesRegex(CommandError, "pager not found"):
                self.run_crash_command("set scroll less")

        def neither(cmd):
            if cmd == "less" or cmd == "more":
                return None
            else:
                return unittest.mock.DEFAULT

        with unittest.mock.patch(
            "shutil.which", side_effect=neither, wraps=shutil.which
        ):
            self.prog.config.pop("crash_pager", None)
            cmd = self.run_crash_command("set scroll")
            self.assertIn("scroll: off (pager not found)", cmd.stdout)

            with self.assertRaisesRegex(CommandError, "pager not found"):
                self.run_crash_command("set scroll less")
            with self.assertRaisesRegex(CommandError, "pager not found"):
                self.run_crash_command("set scroll more")

    def test_radix(self):
        self.addCleanup(self.prog.config.pop, "crash_radix", None)

        cmd = self.run_crash_command("set radix 10")
        self.assertIn("output radix: 10 (decimal)", cmd.stdout)
        cmd = self.run_crash_command("set radix")
        self.assertIn("output radix: 10 (decimal)", cmd.stdout)

        cmd = self.run_crash_command("set radix 16")
        self.assertIn("output radix: 16 (hex)", cmd.stdout)
        cmd = self.run_crash_command("set radix")
        self.assertIn("output radix: 16 (hex)", cmd.stdout)

    def test_radix_invalid(self):
        self.assertRaisesRegex(
            CommandArgumentError,
            "invalid value for radix",
            self.run_crash_command,
            "set radix 0",
        )
        self.assertRaisesRegex(
            CommandArgumentError,
            "invalid value for radix",
            self.run_crash_command,
            "set radix sixty",
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import os
from pathlib import Path
import tempfile

from drgn import (
    DebugInfoOptions,
    KmodSearchMethod,
    MainModule,
    Program,
    RelocatableModule,
)
from drgn.helpers.linux.module import find_module
from tests import modifyenv
from tests.linux_kernel import LinuxKernelTestCase, skip_unless_have_test_kmod


def iter_proc_modules():
    try:
        f = open("/proc/modules", "r")
    except FileNotFoundError:
        return
    with f:
        for line in f:
            tokens = line.split()
            yield tokens[0], int(tokens[5], 16)


class TestDebugInfo(LinuxKernelTestCase):
    def test_debug_info(self):
        # This is actually two test cases squished into one to avoid indexing
        # vmlinux another time.
        prog = Program()
        prog.set_kernel()
        prog.set_enabled_debug_info_finders([])

        with self.subTest("vmlinux_no_build_id"):
            for module, _ in prog.loaded_modules():
                if isinstance(module, MainModule):
                    module.build_id = None
                    break
            else:
                self.fail("main module not found")
            prog.load_debug_info([self.prog.main_module().debug_file_path])
            self.assertEqual(
                prog.main_module().debug_file_path,
                self.prog.main_module().debug_file_path,
            )

        with self.subTest("kmod_walk"), tempfile.TemporaryDirectory() as temp_dir:
            temp_dir = Path(temp_dir)
            found_modules = set()
            for i, module in enumerate(self.prog.modules()):
                if isinstance(module, RelocatableModule) and module.debug_file_path:
                    found_modules.add(module.name)
                    link = temp_dir / str(i) / (module.name + ".ko")
                    link.parent.mkdir()
                    link.symlink_to(module.debug_file_path)

            modules = [
                module
                for module, _ in prog.loaded_modules()
                if module.name in found_modules
            ]
            prog.find_standard_debug_info(
                modules,
                options=DebugInfoOptions(
                    kernel_directories=(temp_dir,), try_kmod=KmodSearchMethod.WALK
                ),
            )
            for module in modules:
                with self.subTest(module=module.name):
                    self.assertIsNotNone(module.debug_file_path)


class TestModule(LinuxKernelTestCase):
    def test_loaded_modules(self):
        expected = [("kernel", None), *iter_proc_modules()]

        loaded_modules = []
        for module, _ in self.prog.loaded_modules():
            if isinstance(module, RelocatableModule):
                loaded_modules.append((module.name, module.address))
            else:
                loaded_modules.append((module.name, None))

        self.assertCountEqual(loaded_modules, expected)

    @skip_unless_have_test_kmod
    def test_find(self):
        self.assertEqual(self.prog.main_module().name, "kernel")
        for name, address in iter_proc_modules():
            if name == "drgn_test":
                self.assertEqual(
                    self.prog.relocatable_module(name, address).name, "drgn_test"
                )
                break
        else:
            self.fail("test module not found")

    @skip_unless_have_test_kmod
    def test_find_by_obj(self):
        for module in self.prog.modules():
            if module.name == "drgn_test":
                break
        else:
            self.fail("test module not found")

        module_obj = find_module(self.prog, "drgn_test")
        self.assertEqual(self.prog.linux_kernel_loadable_module(module_obj), module)
        self.assertEqual(
            self.prog.linux_kernel_loadable_module(module_obj, create=True), module
        )

    def test_no_sys_module(self):
        # Test that we get the same modules with and without using /sys/module.

        def module_dict(prog):
            return {
                (module.name, module.address): (
                    (
                        None
                        if module.address_ranges is None
                        else sorted(module.address_ranges)
                    ),
                    module.build_id,
                    dict(module.section_addresses),
                )
                for module, _ in prog.loaded_modules()
                if isinstance(module, RelocatableModule)
            }

        use_sys_module = int(os.environ.get("DRGN_USE_SYS_MODULE", "1")) != 0

        with modifyenv({"DRGN_USE_SYS_MODULE": str(int(not use_sys_module))}):
            prog = Program()
            prog.set_kernel()

            if use_sys_module:
                with_sys_module = module_dict(self.prog)
                without_sys_module = module_dict(prog)
            else:
                with_sys_module = module_dict(prog)
                without_sys_module = module_dict(self.prog)

            self.assertEqual(with_sys_module, without_sys_module)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later


import dataclasses
import inspect
import os
from pathlib import Path
import subprocess
from typing import Dict, Mapping, NamedTuple, Optional, Sequence, Union

from _drgn_util.platform import NORMALIZED_MACHINE_NAME
from util import KernelVersion, out_of_date

# Kernel versions that we run tests on and therefore support. Keep this in sync
# with docs/support_matrix.rst.
SUPPORTED_KERNEL_VERSIONS = (
    "7.1",
    "7.0",
    "6.19",
    "6.18",
    "6.17",
    "6.16",
    "6.15",
    "6.14",
    "6.13",
    "6.12",
    "6.11",
    "6.10",
    "6.9",
    "6.8",
    "6.7",
    "6.6",
    "6.5",
    "6.4",
    "6.3",
    "6.2",
    "6.1",
    "6.0",
    "5.19",
    "5.18",
    "5.17",
    "5.16",
    "5.15",
    "5.14",
    "5.13",
    "5.12",
    "5.11",
    "5.10",
    "5.4",
    "4.19",
    "4.14",
    "4.9",
)

KERNEL_ORG_COMPILER_VERSION = "12.5.0"


BASE_KCONFIG = """
CONFIG_EXPERT=y
CONFIG_MODULES=y
CONFIG_MODULE_UNLOAD=y
CONFIG_CC_OPTIMIZE_FOR_SIZE=y

# We run the tests in KVM.
CONFIG_HYPERVISOR_GUEST=y
CONFIG_KVM_GUEST=y
CONFIG_PARAVIRT=y
CONFIG_PARAVIRT_SPINLOCKS=y

# Minimum requirements for vmtest.
CONFIG_9P_FS=y
CONFIG_DEVTMPFS=y
CONFIG_INET=y
CONFIG_NET=y
CONFIG_NETWORK_FILESYSTEMS=y
CONFIG_NET_9P=y
CONFIG_NET_9P_VIRTIO=y
CONFIG_OVERLAY_FS=y
CONFIG_PCI=y
CONFIG_PROC_FS=y
CONFIG_SYSFS=y
CONFIG_TMPFS=y
CONFIG_TMPFS_XATTR=y
CONFIG_VIRTIO_BLK=y
CONFIG_VIRTIO_CONSOLE=y
CONFIG_VIRTIO_PCI=y
CONFIG_HW_RANDOM=m
CONFIG_HW_RANDOM_VIRTIO=m

# Lots of things expect Unix sockets.
CONFIG_UNIX=y

# drgn needs debug info.
CONFIG_DEBUG_KERNEL=y
CONFIG_DEBUG_INFO=y
CONFIG_DEBUG_INFO_DWARF4=y

# For testing live kernel debugging with /proc/kcore.
CONFIG_PROC_KCORE=y
# drgn needs /proc/kallsyms in some cases. Some test cases also need it.
CONFIG_KALLSYMS=y
CONFIG_KALLSYMS_ALL=y

# For testing kernel core dumps with /proc/vmcore.
CONFIG_CRASH_DUMP=y
CONFIG_PROC_VMCORE=y
CONFIG_KEXEC=y
CONFIG_KEXEC_FILE=y
# Needed for CONFIG_KEXEC_FILE.
CONFIG_CRYPTO=y
CONFIG_CRYPTO_SHA256=y

# So that we can trigger a crash with /proc/sysrq-trigger.
CONFIG_MAGIC_SYSRQ=y

# For testing kernel core dumps from QEMU's dump-guest-memory command.
CONFIG_FW_CFG_SYSFS=y

# kmodify breakpoints need kprobes.
CONFIG_KPROBES=y

# For BPF tests.
CONFIG_BPF_SYSCALL=y
CONFIG_BPF_JIT=y
CONFIG_BPF_JIT_ALWAYS_ON=y
CONFIG_CGROUP_BPF=y
CONFIG_DEBUG_INFO_BTF=y
CONFIG_DEBUG_INFO_BTF_MODULES=y

# For cgroup tests.
CONFIG_CGROUPS=y
# To select CONFIG_SOCK_CGROUP_DATA. (CONFIG_CGROUP_BPF also selects
# CONFIG_SOCK_CGROUP_DATA, but that's only present since Linux kernel commit
# 3007098494be ("cgroup: add support for eBPF programs") (in v4.10)).
CONFIG_CGROUP_NET_CLASSID=y

# For ipc tests.
CONFIG_SYSVIPC=y

# For kconfig tests.
CONFIG_IKCONFIG=m
CONFIG_IKCONFIG_PROC=y

# For filesystem tests.
CONFIG_BTRFS_FS=m
# Don't waste time benchmarking in raid6_pq just to load the Btrfs module.
CONFIG_RAID6_PQ_BENCHMARK=n
CONFIG_EXT4_FS=m
CONFIG_XFS_FS=m

# For mm tests.
CONFIG_ANON_VMA_NAME=y
CONFIG_HUGETLBFS=y
CONFIG_MEMORY_HOTPLUG=y
CONFIG_MEMORY_HOTREMOVE=y

# For net tests.
CONFIG_NAMESPACES=y

# For nodemask tests.
CONFIG_NUMA=y

# For sched tests.
CONFIG_CGROUP_SCHED=y
CONFIG_FAIR_GROUP_SCHED=y
CONFIG_SCHEDSTATS=y

# For slab allocator tests.
CONFIG_SLAB_FREELIST_HARDENED=y

# For Traffic Control tests.
CONFIG_NET_SCHED=y
CONFIG_NET_SCH_PRIO=m
CONFIG_NET_SCH_SFQ=m
CONFIG_NET_SCH_TBF=m
CONFIG_NET_SCH_INGRESS=m
CONFIG_NET_CLS_ACT=y
CONFIG_NETDEVICES=y
CONFIG_DUMMY=m

# To enable CONFIG_XARRAY_MULTI for xarray tests.
CONFIG_TRANSPARENT_HUGEPAGE=y
CONFIG_READ_ONLY_THP_FOR_FS=y

# For tools/fsrefs.py tests.
CONFIG_BINFMT_MISC=m
CONFIG_BLK_DEV_LOOP=m
CONFIG_DNOTIFY=y
CONFIG_FANOTIFY=y
CONFIG_INOTIFY_USER=y
CONFIG_PERF_EVENTS=y
CONFIG_SWAP=y
# We need to try two names here because of Linux kernel commit 6b0b7551428e
# ("perf/core: Rename CONFIG_[UK]PROBE_EVENT to CONFIG_[UK]PROBE_EVENTS") (in
# v4.11).
CONFIG_UPROBE_EVENT=y
CONFIG_UPROBE_EVENTS=y
CONFIG_USER_NS=y
"""


@dataclasses.dataclass(frozen=True, eq=False)
class KernelFlavor:
    name: str
    description: str
    config: str


KERNEL_FLAVORS = {
    flavor.name: flavor
    for flavor in (
        KernelFlavor(
            name="default",
            description="Default configuration",
            config="""
                CONFIG_SMP=y
                CONFIG_SLUB=y
                # For slab tests.
                CONFIG_SLUB_DEBUG=y
                CONFIG_RANDOMIZE_BASE=y
            """,
        ),
        KernelFlavor(
            name="alternative",
            description="SLAB allocator, module versioning, no KASLR",
            config="""
                CONFIG_SMP=y
                CONFIG_SLAB=y
                # Linux kernel commit eb07c4f39c3e ("mm/slab: rename
                # CONFIG_SLAB to CONFIG_SLAB_DEPRECATED") (in v6.5) renamed the
                # option for SLAB.
                CONFIG_SLAB_DEPRECATED=y
                # Linux kernel commit 16a1d968358a ("mm/slab: remove mm/slab.c
                # and slab_def.h") (in v6.8) removed SLAB. Test this
                # non-default SLUB option instead.
                CONFIG_SLUB_CPU_PARTIAL=n
                CONFIG_RANDOMIZE_BASE=n

                # Options that require additional metadata for kmodify.
                CONFIG_MEM_ALLOC_PROFILING=y
                CONFIG_MEM_ALLOC_PROFILING_ENABLED_BY_DEFAULT=y
                CONFIG_MODVERSIONS=y
            """,
        ),
        KernelFlavor(
            name="tiny",
            description="no SMP, no PREEMPT, no KASLR, and SLUB_TINY or SLOB allocator",
            config="""
                CONFIG_SMP=n
                CONFIG_SLOB=y
                # Linux kernel commit 149b6fa228ed ("mm, slob: rename CONFIG_SLOB to
                # CONFIG_SLOB_DEPRECATED") (in v6.2) renamed the option for SLOB.
                CONFIG_SLOB_DEPRECATED=y
                # Linux kernel commit c9929f0e344a ("mm/slob: remove
                # CONFIG_SLOB") (in v6.4) removed SLOB. Use SLUB_TINY instead,
                # which was introduced in Linux kernel commit e240e53ae0ab
                # ("mm, slub: add CONFIG_SLUB_TINY") (in v6.2).
                CONFIG_SLUB_TINY=y
                # Cover the case of disabling this feature on kernel versions
                # that support it.
                CONFIG_ANON_VMA_NAME=n
                # CONFIG_PREEMPT_DYNAMIC is not set
                CONFIG_PREEMPT_NONE=y
                # !PREEMPTION && !SMP will also select TINY_RCU.
                CONFIG_RANDOMIZE_BASE=n
            """,
        ),
    )
}


@dataclasses.dataclass(frozen=True, eq=False)
class Architecture:
    # Architecture name. This matches the names used by
    # _drgn_util.platform.NORMALIZED_MACHINE_NAME and qemu-system-$arch_name.
    name: str
    # Value of ARCH variable to build the Linux kernel.
    kernel_arch: str
    # Directory under arch/ in the Linux kernel source tree.
    kernel_srcarch: str
    # Name of the architecture in Debian.
    debian_arch: str
    # GCC target triplet on Debian.
    debian_gcc_target: str
    # Linux kernel configuration options.
    kernel_config: str
    # Flavor-specific Linux kernel configuration options.
    kernel_flavor_configs: Mapping[str, str]
    # Name of compiler host on
    # https://mirrors.kernel.org/pub/tools/crosstool/.
    kernel_org_compiler_host_name: Optional[str]
    # Name of compiler target on
    # https://mirrors.kernel.org/pub/tools/crosstool/.
    kernel_org_compiler_target_name: str
    # Options to pass to QEMU.
    qemu_options: Sequence[str]
    # Console device when using QEMU.
    qemu_console: str
    # Minimum supported kernel version.
    min_kernel_version: Optional[KernelVersion] = None

    def kernel_version_supported(self, version: Union[KernelVersion, str]) -> bool:
        if self.min_kernel_version is None:
            return True
        if not isinstance(version, KernelVersion):
            version = KernelVersion(version)
        return version >= self.min_kernel_version


ARCHITECTURES = {
    arch.name: arch
    for arch in (
        Architecture(
            name="aarch64",
            kernel_arch="arm64",
            kernel_srcarch="arm64",
            debian_arch="arm64",
            debian_gcc_target="aarch64-linux-gnu",
            kernel_config="""
                CONFIG_PCI_HOST_GENERIC=y
                CONFIG_RTC_CLASS=y
                CONFIG_RTC_DRV_PL031=y
                CONFIG_SERIAL_AMBA_PL011=y
                CONFIG_SERIAL_AMBA_PL011_CONSOLE=y
            """,
            kernel_flavor_configs={
                "default": """
                    CONFIG_ARM64_4K_PAGES=y
                """,
                "alternative": """
                    CONFIG_ARM64_64K_PAGES=y
                """,
                "tiny": """
                    CONFIG_ARM64_16K_PAGES=y
                """,
            },
            kernel_org_compiler_host_name="arm64",
            kernel_org_compiler_target_name="aarch64-linux",
            qemu_options=("-M", "virt", "-cpu", "cortex-a76"),
            qemu_console="ttyAMA0",
            # /proc/kcore is broken on older versions.
            min_kernel_version=KernelVersion("4.19"),
        ),
        Architecture(
            name="arm",
            kernel_arch="arm",
            kernel_srcarch="arm",
            debian_arch="armhf",
            debian_gcc_target="arm-linux-gnueabihf",
            kernel_config="""
                CONFIG_NR_CPUS=8
                # Debian armhf userspace assumes EABI and VFP.
                CONFIG_AEABI=y
                CONFIG_VFP=y
                CONFIG_ARCH_VIRT=y
                CONFIG_PCI_HOST_GENERIC=y
                CONFIG_RTC_CLASS=y
                CONFIG_RTC_DRV_PL031=y
                CONFIG_SERIAL_AMBA_PL011=y
                CONFIG_SERIAL_AMBA_PL011_CONSOLE=y
                # Before Linux kernel commit f05eb1d24eb5 ("ARM:
                # stackprotector: prefer compiler for TLS based per-task
                # protector") (in v5.18), this enables the
                # arm_ssp_per_task_plugin GCC plugin, which fails to build with
                # the kernel.org cross compiler.
                CONFIG_STACKPROTECTOR_PER_TASK=n
            """,
            kernel_flavor_configs={
                "default": """
                    CONFIG_VMSPLIT_2G=y
                    CONFIG_HIGHMEM=n
                    CONFIG_ARM_LPAE=n
                """,
                "alternative": """
                    CONFIG_VMSPLIT_2G=y
                    CONFIG_HIGHMEM=n
                    CONFIG_ARM_LPAE=y
                """,
                "tiny": """
                    CONFIG_VMSPLIT_3G=y
                    CONFIG_HIGHMEM=y
                    CONFIG_ARM_LPAE=n
                """,
            },
            kernel_org_compiler_host_name=None,
            kernel_org_compiler_target_name="arm-linux-gnueabi",
            qemu_options=("-M", "virt,highmem=off"),
            qemu_console="ttyAMA0",
            # /proc/kcore is broken on older versions.
            min_kernel_version=KernelVersion("4.19"),
        ),
        Architecture(
            name="ppc64",
            kernel_arch="powerpc",
            kernel_srcarch="powerpc",
            debian_arch="ppc64el",
            debian_gcc_target="powerpc64le-linux-gnu",
            kernel_config="""
                CONFIG_PPC64=y
                CONFIG_CPU_LITTLE_ENDIAN=y
                # Debian ppc64el userspace assumes AltiVec and VSX support.
                CONFIG_ALTIVEC=y
                CONFIG_VSX=y
                CONFIG_RTC_CLASS=y
                CONFIG_RTC_DRV_GENERIC=y
                CONFIG_HVC_CONSOLE=y
                # This has a missing dependency in v6.5-rc1 that causes a build
                # failure, and we don't need it.
                CONFIG_CRYPTO_AES_GCM_P10=n
            """,
            kernel_flavor_configs={},
            kernel_org_compiler_host_name="ppc64le",
            kernel_org_compiler_target_name="powerpc64-linux",
            qemu_options=(),
            qemu_console="hvc0",
            # Need an implementation of
            # linux_kernel_live_direct_mapping_fallback in libdrgn.
            min_kernel_version=KernelVersion("4.11"),
        ),
        Architecture(
            name="riscv64",
            kernel_arch="riscv",
            kernel_srcarch="riscv",
            debian_arch="riscv64",
            debian_gcc_target="riscv64-linux-gnu",
            kernel_config="""
                CONFIG_ARCH_RV64I=y
                CONFIG_KERNEL_UNCOMPRESSED=y
                CONFIG_PCI_HOST_GENERIC=y
                CONFIG_SERIAL_8250=y
                CONFIG_SERIAL_8250_CONSOLE=y
                CONFIG_SERIAL_OF_PLATFORM=y
                # For some reason, tests run significantly slower on a
                # relocatable kernel. Additionally, as of Linux 6.19, the KASLR
                # offset is not in VMCOREINFO:
                # https://lore.kernel.org/linux-riscv/aXiF55y5D49gcpUg@adminpc-PowerEdge-R7525/.
                # So, disable KASLR for now.
                CONFIG_RANDOMIZE_BASE=n
            """,
            kernel_flavor_configs={},
            kernel_org_compiler_host_name=None,
            kernel_org_compiler_target_name="riscv64-linux",
            qemu_options=("-M", "virt"),
            qemu_console="ttyS0",
            # RISC-V support was added in Linux 4.15. Linux 6.11 and older need
            # patches to handle more relocation types in kexec_file_load(), and
            # even after patching, 6.7-6.11 hang on kexec. Linux 6.6 and older
            # need a backport of Linux kernel commit e0c0a7c35f67 ("riscv:
            # select ARCH_PROC_KCORE_TEXT"). Linux 6.2 is the oldest version
            # that I can get to boot in QEMU so far. Linux 5.15 is the oldest
            # version that will build without additional patches.
            min_kernel_version=KernelVersion("6.12"),
        ),
        Architecture(
            name="s390x",
            kernel_arch="s390",
            kernel_srcarch="s390",
            debian_arch="s390x",
            debian_gcc_target="s390x-linux-gnu",
            kernel_config="""
                # Needed for CONFIG_KEXEC_FILE.
                CONFIG_CRYPTO_SHA256_S390=y

                # Before Linux kernel commit e950d1f84d3c ("s390/percpu: Get
                # rid of ARCH_MODULE_NEEDS_WEAK_PER_CPU") (in v6.19), memory
                # allocation profiling triggers a bunch of
                # WARN_ON(tag->counters) warnings on s390x. It's not worth
                # dealing with.
                CONFIG_MEM_ALLOC_PROFILING=n
            """,
            kernel_flavor_configs={},
            kernel_org_compiler_host_name=None,
            kernel_org_compiler_target_name="s390-linux",
            qemu_options=(),
            qemu_console="ttysclp0",
            # Need an implementation of
            # linux_kernel_live_direct_mapping_fallback in libdrgn.
            min_kernel_version=KernelVersion("4.11"),
        ),
        Architecture(
            name="x86_64",
            kernel_arch="x86_64",
            kernel_srcarch="x86",
            debian_arch="amd64",
            debian_gcc_target="x86_64-linux-gnu",
            kernel_config="""
                CONFIG_RTC_CLASS=y
                CONFIG_RTC_DRV_CMOS=y
                CONFIG_SERIAL_8250=y
                CONFIG_SERIAL_8250_CONSOLE=y
            """,
            kernel_flavor_configs={
                "alternative": """
                    CONFIG_UNWINDER_FRAME_POINTER=y
                    # Before Linux kernel commit 11af847446ed ("x86/unwind:
                    # Rename unwinder config options to 'CONFIG_UNWINDER_*'")
                    # (in v4.15).
                    CONFIG_FRAME_POINTER_UNWINDER=y
                """,
            },
            kernel_org_compiler_host_name="x86_64",
            kernel_org_compiler_target_name="x86_64-linux",
            qemu_options=("-nodefaults",),
            qemu_console="ttyS0",
        ),
    )
}


HOST_ARCHITECTURE = ARCHITECTURES.get(NORMALIZED_MACHINE_NAME)


class Kernel(NamedTuple):
    arch: Architecture
    release: str
    path: Path


def local_kernel(arch: Architecture, path: Path) -> Kernel:
    return Kernel(
        arch=arch,
        release=(path / "build/include/config/kernel.release").read_text().strip(),
        path=path,
    )


class Compiler(NamedTuple):
    target: Architecture
    bin: Path
    prefix: str

    def env(self) -> Dict[str, str]:
        path = str(self.bin.resolve())
        try:
            path += ":" + os.environ["PATH"]
        except KeyError:
            pass
        return {
            "PATH": path,
            "CROSS_COMPILE": self.prefix,
        }


def compiler_name(host_name: str, target_name: str) -> str:
    return f"{host_name}-gcc-{KERNEL_ORG_COMPILER_VERSION}-nolibc-{target_name}"


def compiler_url(host_name: str, target_name: str) -> str:
    return f"https://github.com/osandov/drgn/releases/download/vmtest-compilers/{compiler_name(host_name, target_name)}.tar.xz"


def kconfig_localversion(arch: Architecture, flavor: KernelFlavor, version: str) -> str:
    vmtest_kernel_version = [
        # Increment the major version to rebuild every
        # architecture/flavor/version combination.
        40,
        # The minor version makes the default flavor the "latest" version.
        1 if flavor.name == "default" else 0,
    ]
    patch_level = 0
    # If only specific architecture/flavor/version combinations need to be
    # rebuilt, conditionally increment the patch level here.
    if flavor.name == "alternative" and KernelVersion(version) >= KernelVersion("6.10"):
        patch_level += 1
        if arch.name == "s390x":
            patch_level += 1
    if patch_level:
        vmtest_kernel_version.append(patch_level)

    return "-vmtest" + ".".join(str(n) for n in vmtest_kernel_version) + flavor.name


def kconfig(arch: Architecture, flavor: KernelFlavor) -> str:
    return f"""\
# Minimal Linux kernel configuration for booting into vmtest and running drgn
# tests ({arch.name} {flavor.name} flavor).

CONFIG_LOCALVERSION=""
CONFIG_LOCALVERSION_AUTO=n

# base options

{inspect.cleandoc(BASE_KCONFIG)}

# {flavor.name} flavor options

{inspect.cleandoc(flavor.config)}

# {arch.name} options

{inspect.cleandoc(arch.kernel_config)}

# {arch.name} {flavor.name} flavor options

{inspect.cleandoc(arch.kernel_flavor_configs.get(flavor.name, ""))}
"""


def _run_autoreconf() -> None:
    if out_of_date(
        "libdrgn/Makefile.in", "libdrgn/Makefile.am", "libdrgn/configure.ac"
    ) or out_of_date("libdrgn/configure", "libdrgn/configure.ac"):
        try:
            subprocess.check_call(["autoreconf", "-i", "libdrgn"])
        except BaseException:
            try:
                os.remove("libdrgn/configure")
            except FileNotFoundError:
                pass
            try:
                os.remove("libdrgn/Makefile.in")
            except FileNotFoundError:
                pass
            raise



# Copyright (c) 2025, Oracle and/or its affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import ctypes
import math
import os
import tempfile

from drgn import Architecture, FaultError, PlatformFlags
from drgn.helpers.linux.pid import find_task
from tests.linux_kernel import (
    skip_if_highmem,
    skip_unless_have_full_mm_support,
    skip_unless_have_test_kmod,
)
from tests.linux_kernel.crash_commands import CrashCommandTestCase


class TestRd(CrashCommandTestCase):
    def setUp(self):
        self.w = self.prog.address_size() * 2

    def test_no_options(self):
        address = self.prog["init_task"].address_
        cmd = self.check_crash_command(f"rd {address:x}")
        self.assertRegex(cmd.stdout, rf"^{address:{self.w}x}:")
        self.assertRegex(cmd.stdout, r"[0-9a-f]{" + str(self.w) + "}")

    def test_count(self):
        address = self.prog["init_task"].address_
        cmd = self.check_crash_command(f"rd -64 {address:x} 3")
        lines = cmd.stdout.splitlines()
        self.assertGreaterEqual(len(lines), 2)
        self.assertRegex(lines[0], rf"^{address:{self.w}x}:")

    def test_format_decimal(self):
        address = self.prog["init_task"].address_
        cmd = self.check_crash_command(f"rd -d {address:x}")
        self.assertRegex(cmd.stdout, rf"^{address:{self.w}x}:")
        # Should have decimal numbers
        self.assertRegex(cmd.stdout, r"\b-?\d+\b")

    def test_units(self):
        data = b"Linux ve"
        address = self.prog.symbol("linux_banner").address
        if self.prog.platform.flags & PlatformFlags.IS_LITTLE_ENDIAN:
            endian = "little"
        else:
            endian = "big"
        for unit in (1, 2, 4, 8):
            with self.subTest(unit=unit):
                count = 8 // unit
                cmd = self.check_crash_command(f"rd -{unit * 8} linux_banner {count}")
                value_list = []
                for i in range(count):
                    int_bytes = data[i * unit : (i + 1) * unit]
                    value_list.append(int.from_bytes(int_bytes, byteorder=endian))
                values = " ".join(f"{v:0{unit * 2}x}" for v in value_list)
                self.assertRegex(
                    cmd.stdout,
                    rf"^{address:{self.w}x}:  {values} +Linux ve$",
                )

    def test_network_byte_order(self):
        data = b"Linux ve"
        address = self.prog.symbol("linux_banner").address
        for unit in (2, 4, 8):
            with self.subTest(unit=unit):
                count = 8 // unit
                cmd = self.check_crash_command(
                    f"rd -{unit * 8} -N linux_banner {count}"
                )
                value_list = []
                for i in range(count):
                    int_bytes = data[i * unit : (i + 1) * unit]
                    value_list.append(int.from_bytes(int_bytes, byteorder="big"))
                values = " ".join(f"{v:0{unit * 2}x}" for v in value_list)
                self.assertRegex(
                    cmd.stdout,
                    rf"^{address:{self.w}x}:  {values} +Linux ve$",
                )

    def test_ascii(self):
        cmd = self.check_crash_command("rd -a linux_banner")
        address = self.prog.symbol("linux_banner").address
        self.assertRegex(cmd.stdout, rf"{address:{self.w}x}:  Linux version")
        lines = cmd.stdout.strip().split("\n")

        # The ascii variant formats output over multiple lines, breaking at 79
        # characters. The linux banner is good test data for this, because it
        # has no special whitespace and it is generally rather long. While we
        # don't necessarily care about the specific line break point, it is good
        # to verify that we get roughly the expected amount of output, as a
        # smoke test.
        string = self.prog["linux_banner"].string_()
        chars_per_line = 79 - self.w - 3
        line_count = math.ceil(len(string) / chars_per_line)
        self.assertEqual(len(lines), line_count)

    def test_ascii_count(self):
        cmd = self.check_crash_command("rd -a linux_banner 5")
        address = self.prog.symbol("linux_banner").address
        self.assertEqual(cmd.stdout.strip(), f"{address:x}:  Linux")

    def test_ascii_offset(self):
        address = self.prog.symbol("linux_banner").address
        cmd = self.check_crash_command("rd -o 4 -a linux_banner")
        self.assertRegex(cmd.stdout, f"^{address + 4:{self.w}x}:  x version.*")

    def test_symbol(self):
        address = self.prog["init_task"].address_
        cmd = self.check_crash_command("rd init_task")
        self.assertRegex(cmd.stdout, rf"^{address:{self.w}x}:")

    def test_annotate_symbols(self):
        address = self.prog["slab_caches"].prev.value_()
        cmd = self.check_crash_command(f"rd -s {address:x} 2")
        self.assertRegex(cmd.stdout, rf"^{address:{self.w}x}:  slab_caches\+0")

    @skip_unless_have_full_mm_support
    @skip_unless_have_test_kmod
    def test_annotate_slab(self):
        address = self.prog.symbol("drgn_test_small_slab_objects").address
        cmd = self.check_crash_command("rd -S drgn_test_small_slab_objects 2")
        if self.prog["drgn_test_slob"]:
            self.assertRegex(
                cmd.stdout,
                rf"^{address:{self.w}x}:  \[unknown slab object\] +\[unknown slab object\]",
            )
        else:
            self.assertRegex(
                cmd.stdout,
                rf"^{address:{self.w}x}:  \[drgn_test_small\] +\[drgn_test_small\]",
            )

    @skip_unless_have_full_mm_support
    @skip_unless_have_test_kmod
    def test_annotate_slab_verbose(self):
        address = self.prog.symbol("drgn_test_small_slab_objects").address
        cmd = self.check_crash_command("rd -SS drgn_test_small_slab_objects 2")
        if self.prog["drgn_test_slob"]:
            self.assertRegex(
                cmd.stdout,
                rf"^{address:{self.w}x}:  \[[0-9a-f]+:unknown slab object\] +\[[0-9a-f]+:unknown slab object\]",
            )
        else:
            self.assertRegex(
                cmd.stdout,
                rf"^{address:{self.w}x}:  \[[0-9a-f]+:drgn_test_small\] +\[[0-9a-f]+:drgn_test_small\]",
            )

    def test_reverse(self):
        address = self.prog["init_task"].address_
        cmd = self.check_crash_command(f"rd -R {address:x} 2")
        unit = self.prog.address_size()
        start_addr = address - unit
        self.assertRegex(cmd.stdout, rf"^{start_addr:{self.w}x}:  [0-9a-f]+ [0-9a-f]+")

    def test_raw(self):
        address = self.prog["init_task"].address_
        with tempfile.NamedTemporaryFile(delete=False) as f:
            temp_file = f.name
            cmd = self.check_crash_command(f"rd -r {temp_file} {address:x} 16")
            self.assertIn("16 bytes copied", cmd.stdout)
            self.assertTrue(os.path.exists(temp_file))
            with open(temp_file, "rb") as f:
                data = f.read()
                self.assertEqual(len(data), 16)

    @skip_if_highmem
    def test_ascii_user(self):
        string = "hello, world!"
        buf = ctypes.create_string_buffer(string.encode("ascii"))
        address = ctypes.addressof(buf)
        self.prog.config["crash_context"] = find_task(self.prog, os.getpid())

        # On s390x, the user and kernel address spaces occupy the same range, so
        # we cannot detect whether an address is a user address. Explicitly pass
        # the hint here.
        need_hint = self.prog.platform.arch == Architecture.S390X

        expected_value = f"{address:{self.w}x}:  hello world, I am a string!"

        if not need_hint:
            with self.subTest("without -u hint"):
                cmd = self.check_crash_command(f"rd -a -o {self.w} 0x{address:x}")
                self.assertEqual(cmd.stdout.strip(), expected_value)
                self.assertIn("access_process_vm", cmd.drgn_option)
            self.assertNotIn("prog.read", cmd.drgn_option)
        with self.subTest("with -u hint"):
            cmd = self.check_crash_command(f"rd -u -a -o {self.w} 0x{address:x}")
            self.assertEqual(cmd.stdout.strip(), expected_value)
            self.assertIn("access_process_vm", cmd.drgn_option)
            self.assertNotIn("prog.read", cmd.drgn_option)

    def check_failing_command(self, command, exc_type):
        # Directly running the command should fail
        with self.assertRaises(exc_type):
            self.run_crash_command(command)

        # The drgn option should generate valid code that compiles
        self.run_crash_command_drgn_option(command, mode="compile")

        # But the code should also fail with the same error as the command
        with self.assertRaises(exc_type):
            self.run_crash_command_drgn_option(command, mode="exec")

    def test_not_existing_symbol(self):
        self.check_failing_command("rd this_symbol_does_not_exist", LookupError)

    @skip_unless_have_full_mm_support
    def test_fault(self):
        self.check_failing_command("rd -u 0", FaultError)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

from drgn import DebugInfoOptions, KmodSearchMethod, Program
from tests import TestCase


class TestDebugInfoOptions(TestCase):
    def test_list_default(self):
        self.assertEqual(DebugInfoOptions().directories, ("/usr/lib/debug",))

    def test_list_init(self):
        self.assertEqual(
            DebugInfoOptions(directories=["foo", "bar"]).directories, ("foo", "bar")
        )
        self.assertRaises(TypeError, DebugInfoOptions, directories=None)

    def test_list_copy(self):
        self.assertEqual(
            DebugInfoOptions(DebugInfoOptions(directories=["foo", "bar"])).directories,
            ("foo", "bar"),
        )

    def test_list_set(self):
        options = DebugInfoOptions()
        options.directories = ("foo", "bar")
        self.assertEqual(options.directories, ("foo", "bar"))
        with self.assertRaises(TypeError):
            DebugInfoOptions().directories = None

    def test_directories_empty_string(self):
        with self.assertRaises(ValueError):
            DebugInfoOptions().directories = ("",)

    def test_bool_default(self):
        self.assertIs(DebugInfoOptions().try_build_id, True)

    def test_bool_init(self):
        self.assertIs(DebugInfoOptions(try_build_id=False).try_build_id, False)

    def test_bool_copy(self):
        self.assertIs(
            DebugInfoOptions(DebugInfoOptions(try_build_id=False)).try_build_id, False
        )

    def test_bool_set(self):
        options = DebugInfoOptions()
        options.try_build_id = False
        self.assertIs(options.try_build_id, False)

    def test_enum_default(self):
        self.assertEqual(DebugInfoOptions().try_kmod, KmodSearchMethod.DEPMOD_OR_WALK)

    def test_enum_init(self):
        self.assertEqual(
            DebugInfoOptions(try_kmod=KmodSearchMethod.WALK).try_kmod,
            KmodSearchMethod.WALK,
        )
        self.assertRaises(TypeError, DebugInfoOptions, try_kmod=False)

    def test_enum_copy(self):
        self.assertEqual(
            DebugInfoOptions(
                DebugInfoOptions(try_kmod=KmodSearchMethod.DEPMOD)
            ).try_kmod,
            KmodSearchMethod.DEPMOD,
        )

    def test_enum_set(self):
        options = DebugInfoOptions()
        options.try_kmod = KmodSearchMethod.DEPMOD_AND_WALK
        self.assertEqual(options.try_kmod, KmodSearchMethod.DEPMOD_AND_WALK)
        with self.assertRaises(TypeError):
            options.try_kmod = False

    def test_del(self):
        with self.assertRaises(AttributeError):
            del DebugInfoOptions().directories

    def test_repr(self):
        self.assertIn("directories=()", repr(DebugInfoOptions(directories=())))


class TestProgramDebugInfoOptions(TestCase):
    def test_default(self):
        self.assertEqual(
            Program().debug_info_options.directories, DebugInfoOptions().directories
        )

    def test_assign(self):
        prog = Program()
        prog.debug_info_options.directories = ("foo", "bar")
        prog.debug_info_options = DebugInfoOptions(directories=("bar", "baz"))
        self.assertEqual(prog.debug_info_options.directories, ("bar", "baz"))

    def test_assign_list(self):
        prog = Program()
        prog.debug_info_options.directories = ("bar", "foo")
        self.assertEqual(prog.debug_info_options.directories, ("bar", "foo"))



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import os
import resource

from drgn.helpers.linux.pid import find_task
from drgn.helpers.linux.resource import Rlimit, task_rlimits
from tests.linux_kernel import LinuxKernelTestCase


class TestResource(LinuxKernelTestCase):
    def test_task_rlimits(self):
        expected = {}
        for name in (
            "CPU",
            "FSIZE",
            "DATA",
            "STACK",
            "CORE",
            "RSS",
            "NPROC",
            "NOFILE",
            "MEMLOCK",
            "AS",
            "LOCKS",
            "SIGPENDING",
            "MSGQUEUE",
            "NICE",
            "RTPRIO",
            "RTTIME",
        ):
            try:
                r = getattr(resource, "RLIMIT_" + name)
            except AttributeError:
                # RLIMIT_LOCKS was functional only briefly between Linux 2.4.0
                # and Linux 2.4.24, so the resource module doesn't define it.
                if name == "LOCKS":
                    r = 10
                else:
                    raise
            soft, hard = resource.getrlimit(r)
            expected[name] = Rlimit(
                None if soft == resource.RLIM_INFINITY else soft,
                None if hard == resource.RLIM_INFINITY else hard,
            )
        self.assertEqual(
            task_rlimits(find_task(self.prog, os.getpid())),
            expected,
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import collections
import os
from pathlib import Path
import re
import threading

from drgn import Object
from drgn.commands._crash.common import _CRASH_FOREACH_SUBCOMMANDS
from drgn.helpers.common.format import double_quote_ascii_string
from drgn.helpers.linux.mm import TaskRss
from drgn.helpers.linux.pid import find_task
from tests.linux_kernel import (
    fork_and_stop,
    online_cpus,
    skip_if_highmem,
    skip_unless_have_full_mm_support,
)
from tests.linux_kernel.crash_commands import CrashCommandTestCase


def _quoted_comm(pid="self"):
    return double_quote_ascii_string(
        Path(f"/proc/{pid}/comm").read_bytes().rstrip(b"\n")
    )


class TestPs(CrashCommandTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        cls.condition = threading.Condition()
        cls.threads = [threading.Thread(target=cls.thread_func) for _ in range(2)]
        cls.addClassCleanup(cls.cleanup_threads)
        for i, thread in enumerate(cls.threads):
            name = f"drgn_test_ps{i}"
            thread.name = name
            thread.start()

            # Since Python 3.14, the thread name will be set at the OS level,
            # but before that, we have to do it manually.
            comm_path = Path(f"/proc/{thread.native_id}/comm")
            if comm_path.read_text() != name + "\n":
                comm_path.write_text(name + "\n")

    @classmethod
    def thread_func(cls):
        with cls.condition:
            cls.condition.wait()

    @classmethod
    def cleanup_threads(cls):
        with cls.condition:
            cls.condition.notify_all()
        for thread in cls.threads:
            thread.join()

    def _test_drgn_common(self, cmd):
        for variable in (
            "task",
            "pid",
            "ppid",
            "comm",
        ):
            with self.subTest(variable=variable):
                self.assertIsInstance(cmd.drgn_option.globals[variable], Object)
        self.assertIsInstance(cmd.drgn_option.globals["cpu"], int)
        self.assertIsInstance(cmd.drgn_option.globals["state"], str)
        self.assertIsInstance(cmd.drgn_option.globals["mem_usage"], float)
        self.assertIsInstance(cmd.drgn_option.globals["vsize"], int)
        self.assertIsInstance(cmd.drgn_option.globals["rss"], TaskRss)

    def test_no_args(self):
        cmd = self.check_crash_command("ps")

        self.assertRegex(cmd.stdout, r"^\s*PID\s+PPID")
        for pid in [0, 1, os.getpid(), *(thread.native_id for thread in self.threads)]:
            self.assertRegex(cmd.stdout, rf"(?m)^>?\s*{pid}\b")

        self.assertIn("for_each_task(idle=True)", cmd.drgn_option.stdout)
        self.assertNotIn("continue", cmd.drgn_option.stdout)
        self._test_drgn_common(cmd)

    def test_pid(self):
        cmd = self.check_crash_command("ps -H 1")
        foreach_cmd = self.check_crash_command("foreach 1 ps -H", mode="capture")

        for c in (cmd, foreach_cmd):
            self.assertRegex(c.stdout, r"^>?\s*1\b.*\n$")

        self.assertIn("pid = 1", cmd.drgn_option.stdout)
        self.assertIn("find_task(pid)", cmd.drgn_option.stdout)
        self.assertNotIn("for_each_task", cmd.drgn_option.stdout)
        self._test_drgn_common(cmd)

        self.assertEqual(cmd.drgn_option.stdout, foreach_cmd.drgn_option.stdout)

    def test_pid_0(self):
        cmd = self.check_crash_command("ps -H 0")

        for line in cmd.stdout.splitlines(keepends=True):
            self.assertRegex(line, r"^>?\s*0\b.*\n$")

        self.assertIn("for_each_online_cpu(", cmd.drgn_option.stdout)
        self.assertIn("idle_task(", cmd.drgn_option.stdout)
        self.assertNotIn("for_each_task", cmd.drgn_option.stdout)
        self._test_drgn_common(cmd)

    def test_multiple_pids(self):
        cmd = self.check_crash_command(f"ps -H {os.getpid()} 1")

        self.assertRegex(cmd.stdout, rf"^>?\s*1\b.*\n>?\s*{os.getpid()}\b.*\n$")

        self.assertIn("pid = 1\n", cmd.drgn_option.stdout)
        self.assertIn(f"pid = {os.getpid()}\n", cmd.drgn_option.stdout)
        self.assertIn("find_task(pid)", cmd.drgn_option.stdout)
        self.assertNotIn("for_each_task", cmd.drgn_option.stdout)
        self._test_drgn_common(cmd)

    def test_invalid_pid(self):
        cmd = self.check_crash_command("ps -H 2147483647")

        self.assertIn("no such process with PID 2147483647", cmd.stdout)

    def test_task_struct(self):
        address = find_task(self.prog, 1).value_()
        cmd = self.check_crash_command(f"ps -H {address:#x}")

        self.assertRegex(cmd.stdout, r"^>?\s*1\b.*\n$")

        self.assertNotIn("for_each_task", cmd.drgn_option.stdout)
        self.assertIn(hex(address), cmd.drgn_option.stdout)
        self.assertIn('Object(prog, "struct task_struct *", ', cmd.drgn_option.stdout)
        self._test_drgn_common(cmd)

    def test_multiple_task_structs(self):
        cmd = self.check_crash_command(
            f"ps -H {find_task(self.prog, 1).value_():#x} {self.prog['init_task'].address_:#x}"
        )

        self.assertRegex(cmd.stdout, r"^>?\s*0\b.*\n>?\s*1\b.*\n$")

        self.assertNotIn("for_each_task", cmd.drgn_option.stdout)
        self.assertIn('Object(prog, "struct task_struct *", ', cmd.drgn_option.stdout)
        self._test_drgn_common(cmd)

    def test_invalid_task_struct(self):
        address = self.prog["init_task"].address_
        cmd = self.run_crash_command(f"ps -H {address + 1:#x}")

        self.assertIn(f"invalid task_struct: {address + 1:#x}", cmd.stdout)

    def test_pid_and_task_struct(self):
        address = self.prog["init_task"].address_
        cmd = self.check_crash_command(f"ps -H 1 {address:#x}")

        self.assertRegex(cmd.stdout, r"(?m)^>?\s*0\b")
        self.assertRegex(cmd.stdout, r"(?m)^>?\s*1\b")

        self.assertNotIn("for_each_task", cmd.drgn_option.stdout)
        self.assertIn("pid = 1", cmd.drgn_option.stdout)
        self.assertIn("find_task(pid)", cmd.drgn_option.stdout)
        self.assertIn(hex(address), cmd.drgn_option.stdout)
        self.assertIn('Object(prog, "struct task_struct *", ', cmd.drgn_option.stdout)
        self._test_drgn_common(cmd)

    def test_command_names(self):
        self_comm = Path("/proc/self/comm").read_bytes().rstrip(b"\n")
        similar_comm = self_comm + b"1" if len(self_comm) < 15 else self_comm[:-1]
        with fork_and_stop(Path("/proc/self/comm").write_bytes, similar_comm) as (
            pid,
            _,
        ):
            cmd = self.check_crash_command(
                f"ps -H {double_quote_ascii_string(self_comm)} {_quoted_comm(1)}"
            )

        self.assertRegex(cmd.stdout, r"(?m)^>?\s*1\b")
        self.assertRegex(cmd.stdout, rf"(?m)^>?\s*{os.getpid()}\b")
        self.assertNotRegex(cmd.stdout, rf"(?m)^>?\s*{pid}\b")

        self.assertIn("for_each_task(idle=True)", cmd.drgn_option.stdout)
        self.assertIn("comm_string not in ", cmd.drgn_option.stdout)
        self.assertIn("continue", cmd.drgn_option.stdout)
        self._test_drgn_common(cmd)

    def test_backslash_command_name(self):
        with fork_and_stop(Path("/proc/self/comm").write_bytes, b"drgn_test_ps") as (
            pid,
            _,
        ):
            cmd = self.check_crash_command(r"ps -H \drgn_test_ps")

        self.assertRegex(cmd.stdout, rf"(?m)^>?\s*{pid}\b")

        self.assertIn("for_each_task(idle=True)", cmd.drgn_option.stdout)
        self.assertIn('comm_string != "drgn_test_ps"', cmd.drgn_option.stdout)
        self.assertIn("continue", cmd.drgn_option.stdout)
        self._test_drgn_common(cmd)

    def test_command_name_regex(self):
        cmd = self.check_crash_command(r"ps -H 'drgn_test_ps[0-9]'")

        for pid in [thread.native_id for thread in self.threads]:
            self.assertRegex(cmd.stdout, rf"(?m)^>?\s*{pid}\b")

        self.assertIn("for_each_task(idle=True)", cmd.drgn_option.stdout)
        self.assertIn(
            're.search("drgn_test_ps[0-9]", comm_string)', cmd.drgn_option.stdout
        )
        self.assertIn("continue", cmd.drgn_option.stdout)
        self._test_drgn_common(cmd)

    def test_pid_and_command_name(self):
        cmd = self.check_crash_command(f"ps -H {_quoted_comm()} 1")

        self.assertRegex(cmd.stdout, r"(?m)^>?\s*1\b")
        self.assertRegex(cmd.stdout, rf"(?m)^>?\s*{os.getpid()}\b")

        self.assertIn("for_each_task(idle=True)", cmd.drgn_option.stdout)
        self.assertNotIn("find_task(", cmd.drgn_option.stdout)
        self.assertIn("task.pid.value_() == 1", cmd.drgn_option.stdout)
        self.assertIn("comm_string != ", cmd.drgn_option.stdout)
        self.assertIn("continue", cmd.drgn_option.stdout)
        self._test_drgn_common(cmd)

    def test_pid_and_task_struct_and_command_name(self):
        address = self.prog["init_task"].address_
        cmd = self.check_crash_command(f"ps -H {_quoted_comm()} 1 {address:#x}")

        self.assertRegex(cmd.stdout, r"(?m)^>?\s*0\b")
        self.assertRegex(cmd.stdout, r"(?m)^>?\s*1\b")
        self.assertRegex(cmd.stdout, rf"(?m)^>?\s*{os.getpid()}\b")

        self.assertIn("for_each_task(idle=True)", cmd.drgn_option.stdout)
        self.assertNotIn("find_task(", cmd.drgn_option.stdout)
        self.assertIn("task.pid.value_() == 1", cmd.drgn_option.stdout)
        self.assertIn(f"task.value_() == {address:#x}", cmd.drgn_option.stdout)
        self.assertIn("comm_string != ", cmd.drgn_option.stdout)
        self.assertIn("continue", cmd.drgn_option.stdout)
        self._test_drgn_common(cmd)

    def test_kernel(self):
        cmd = self.check_crash_command("ps -k")

        self.assertRegex(cmd.stdout, r"(?m)^>?\s*0\b")
        self.assertNotRegex(cmd.stdout, r"(?m)^>?\s*1\b")

        self.assertIn("for_each_task(idle=True)", cmd.drgn_option.stdout)
        self.assertIn("if not task_is_kthread(task):", cmd.drgn_option.stdout)
        self.assertIn("continue", cmd.drgn_option.stdout)
        self._test_drgn_common(cmd)

    def test_kernel_and_pid(self):
        cmd = self.check_crash_command("ps -H -k 0 2")

        self.assertRegex(cmd.stdout, r"(?:^>?\s*[02]\b.*\n)+")

        self.assertIn("for_each_online_cpu(", cmd.drgn_option.stdout)
        self.assertIn("idle_task(", cmd.drgn_option.stdout)
        self.assertIn("pid = 2", cmd.drgn_option.stdout)
        self.assertIn("find_task(pid)", cmd.drgn_option.stdout)
        self.assertNotIn("for_each_task", cmd.drgn_option.stdout)
        self.assertIn("if not task_is_kthread(task):", cmd.drgn_option.stdout)
        self.assertIn("continue", cmd.drgn_option.stdout)
        self._test_drgn_common(cmd)

    def test_kernel_and_user_pid(self):
        cmd = self.check_crash_command("ps -H -k 1")

        self.assertFalse(cmd.stdout)
        self.assertNotIn("ppid", cmd.drgn_option.globals)

    def test_user(self):
        cmd = self.check_crash_command("ps -u")

        self.assertRegex(cmd.stdout, r"(?m)^>?\s*1\b")
        self.assertNotRegex(cmd.stdout, r"(?m)^>?\s*0\b")

        self.assertIn("for_each_task()", cmd.drgn_option.stdout)
        self.assertIn("if task_is_kthread(task):", cmd.drgn_option.stdout)
        self.assertIn("continue", cmd.drgn_option.stdout)
        self._test_drgn_common(cmd)

    def test_user_and_pid(self):
        cmd = self.check_crash_command("ps -H -u 1")

        self.assertRegex(cmd.stdout, r"^>?\s*1\b.*\n$")

        self.assertIn("pid = 1", cmd.drgn_option.stdout)
        self.assertIn("find_task(pid)", cmd.drgn_option.stdout)
        self.assertNotIn("for_each_task", cmd.drgn_option.stdout)
        self.assertIn("if task_is_kthread(task):", cmd.drgn_option.stdout)
        self.assertIn("continue", cmd.drgn_option.stdout)
        self._test_drgn_common(cmd)

    def test_user_and_kernel_pid(self):
        cmd = self.check_crash_command("ps -H -u 0 2")

        self.assertFalse(cmd.stdout)
        self.assertNotIn("ppid", cmd.drgn_option.globals)

    def test_group_leader(self):
        cmd = self.check_crash_command("ps -G")

        self.assertRegex(cmd.stdout, rf"(?m)^>?\s*{os.getpid()}\b")
        self.assertNotRegex(cmd.stdout, rf"(?m)^>?\s*{self.threads[0].native_id}\b")

        self.assertIn("for_each_task(idle=True)", cmd.drgn_option.stdout)
        self.assertIn("if not thread_group_leader(task):", cmd.drgn_option.stdout)
        self.assertIn("continue", cmd.drgn_option.stdout)
        self._test_drgn_common(cmd)

    def test_group_leader_and_pid(self):
        cmd = self.check_crash_command(f"ps -H -G {os.getpid()}")
        foreach_cmd = self.check_crash_command(
            f"foreach {os.getpid()} ps -H -G", mode="capture"
        )

        for c in (cmd, foreach_cmd):
            self.assertRegex(c.stdout, rf"^>?\s*{os.getpid()}\b.*$")

        self.assertIn("find_task(pid)", cmd.drgn_option.stdout)
        self.assertNotIn("for_each_task", cmd.drgn_option.stdout)
        self.assertIn("if not thread_group_leader(task):", cmd.drgn_option.stdout)
        self.assertIn("task = task.group_leader", cmd.drgn_option.stdout)
        self.assertNotIn("continue", cmd.drgn_option.stdout)
        self._test_drgn_common(cmd)

        self.assertEqual(cmd.drgn_option.stdout, foreach_cmd.drgn_option.stdout)

    def test_group_leader_and_tid(self):
        cmd = self.check_crash_command(f"ps -H -G {self.threads[0].native_id}")

        self.assertRegex(cmd.stdout, rf"^>?\s*{os.getpid()}\b.*$")

        self.assertIn("find_task(pid)", cmd.drgn_option.stdout)
        self.assertNotIn("for_each_task", cmd.drgn_option.stdout)
        self.assertIn("if not thread_group_leader(task):", cmd.drgn_option.stdout)
        self.assertIn("task = task.group_leader", cmd.drgn_option.stdout)
        self.assertNotIn("continue", cmd.drgn_option.stdout)
        self._test_drgn_common(cmd)

    def test_group_leader_and_task_struct(self):
        address = find_task(self.prog, os.getpid()).value_()
        cmd = self.check_crash_command(f"ps -H -G {address:#x}")

        self.assertRegex(cmd.stdout, rf"^>?\s*{os.getpid()}\b.*$")

        self.assertNotIn("for_each_task", cmd.drgn_option.stdout)
        self.assertIn(hex(address), cmd.drgn_option.stdout)
        self.assertIn('Object(prog, "struct task_struct *", ', cmd.drgn_option.stdout)
        self.assertIn("if not thread_group_leader(task):", cmd.drgn_option.stdout)
        self.assertIn("task = task.group_leader", cmd.drgn_option.stdout)
        self._test_drgn_common(cmd)

    def test_group_leader_and_thread_task_struct(self):
        address = find_task(self.prog, self.threads[0].native_id).value_()
        cmd = self.check_crash_command(f"ps -H -G {address:#x}")

        self.assertRegex(cmd.stdout, rf"^>?\s*{os.getpid()}\b.*$")

        self.assertNotIn("for_each_task", cmd.drgn_option.stdout)
        self.assertIn(hex(address), cmd.drgn_option.stdout)
        self.assertIn('Object(prog, "struct task_struct *", ', cmd.drgn_option.stdout)
        self.assertIn("if not thread_group_leader(task):", cmd.drgn_option.stdout)
        self.assertIn("task = task.group_leader", cmd.drgn_option.stdout)
        self._test_drgn_common(cmd)

    def test_group_leader_and_command_name(self):
        cmd = self.check_crash_command(
            f"ps -H -G {_quoted_comm()} {_quoted_comm(self.threads[0].native_id)}"
        )

        self.assertRegex(cmd.stdout, rf"(?m)^>?\s*{os.getpid()}\b")
        self.assertNotRegex(cmd.stdout, rf"(?m)^>?\s*{self.threads[0].native_id}\b")

        self.assertIn("for_each_task(idle=True)", cmd.drgn_option.stdout)
        self.assertIn("if not thread_group_leader(task):", cmd.drgn_option.stdout)
        self.assertNotIn("task = task.group_leader", cmd.drgn_option.stdout)
        self.assertIn("comm_string not in ", cmd.drgn_option.stdout)
        self.assertIn("continue", cmd.drgn_option.stdout)
        self._test_drgn_common(cmd)

    def test_policy(self):
        with fork_and_stop() as pid1, fork_and_stop() as pid2:
            os.sched_setscheduler(pid1, os.SCHED_OTHER, os.sched_param(0))
            os.sched_setscheduler(pid2, os.SCHED_BATCH, os.sched_param(0))

            cmd = self.check_crash_command("ps -y BATCH")
            foreach_cmd = self.check_crash_command(
                "foreach ps -y BATCH", mode="capture"
            )

        for c in (cmd, foreach_cmd):
            self.assertRegex(c.stdout, rf"(?m)^>?\s*{pid2}\b")
            self.assertNotRegex(c.stdout, rf"(?m)^>?\s*{pid1}\b")

        self.assertIn("for_each_task(idle=True)", cmd.drgn_option.stdout)
        self.assertIn("if task.policy.value_() != ", cmd.drgn_option.stdout)
        self.assertIn("continue", cmd.drgn_option.stdout)
        self._test_drgn_common(cmd)

        self.assertEqual(cmd.drgn_option.stdout, foreach_cmd.drgn_option.stdout)

    def test_state(self):
        with fork_and_stop() as pid:
            cmd = self.check_crash_command("foreach R ps")

        self.assertNotRegex(cmd.stdout, rf"(?m)^>?\s*{pid}\b")

        self.assertIn("for_each_task(idle=True)", cmd.drgn_option.stdout)
        self.assertIn("if task_state_to_char(task) != ", cmd.drgn_option.stdout)
        self.assertIn("continue", cmd.drgn_option.stdout)
        self._test_drgn_common(cmd)

    def test_active(self):
        with fork_and_stop() as pid:
            cmd = self.check_crash_command("ps -A")
            foreach_cmd = self.check_crash_command("foreach active ps", mode="capture")

        for c in (cmd, foreach_cmd):
            self.assertNotRegex(c.stdout, rf"(?m)^>?\s*{pid}\b")

        self.assertIn("for_each_task(idle=True)", cmd.drgn_option.stdout)
        self.assertIn("if not task_on_cpu(task)", cmd.drgn_option.stdout)
        self.assertIn("continue", cmd.drgn_option.stdout)
        self._test_drgn_common(cmd)

        self.assertEqual(cmd.drgn_option.stdout, foreach_cmd.drgn_option.stdout)

    def _test_drgn_task_header(self, cmd):
        for variable in (
            "task",
            "pid",
            "command",
        ):
            with self.subTest(variable=variable):
                self.assertIsInstance(cmd.drgn_option.globals[variable], Object)
        self.assertIsInstance(cmd.drgn_option.globals["cpu"], int)

    def test_parents(self):
        with fork_and_stop() as pid:
            parents = [pid]
            current_pid = pid
            while current_pid != 0:
                ppid = int(
                    re.search(
                        r"^PPid:\s*([0-9]+)",
                        Path(f"/proc/{current_pid}/status").read_text(),
                        flags=re.M,
                    ).group(1)
                )
                parents.append(ppid)
                current_pid = ppid

            cmd = self.check_crash_command(f"ps -p {pid}")
            foreach_cmd = self.check_crash_command(
                f"foreach {pid} ps -p", mode="capture"
            )

        regex = ["^"]
        for i, expected_pid in enumerate(reversed(parents)):
            regex.append(" " * i + rf"PID: {expected_pid}\b.*\n")
        regex.append("$")
        for c in (cmd, foreach_cmd):
            self.assertRegex(c.stdout, "".join(regex))

        self.assertIn("task.parent", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["parent"].type_.type_name(), "struct task_struct *"
        )
        self._test_drgn_task_header(cmd)

        self.assertEqual(cmd.drgn_option.stdout, foreach_cmd.drgn_option.stdout)

    def test_children(self):
        with fork_and_stop() as pid1, fork_and_stop() as pid2:
            cmd = self.check_crash_command(f"ps -c {os.getpid()}")
            foreach_cmd = self.check_crash_command(
                f"foreach {os.getpid()} ps -c", mode="capture"
            )

        for c in (cmd, foreach_cmd):
            self.assertRegex(c.stdout, rf"^PID: {os.getpid()}\b")
            for pid in pid1, pid2:
                self.assertRegex(c.stdout, rf"(?m)^  PID: {pid}\b")

        self.assertIn("list_for_each_entry(", cmd.drgn_option.stdout)
        self.assertIn("task.children", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["child"].type_.type_name(), "struct task_struct *"
        )
        self._test_drgn_task_header(cmd)

        self.assertEqual(cmd.drgn_option.stdout, foreach_cmd.drgn_option.stdout)

    def test_no_children(self):
        with fork_and_stop() as pid:
            cmd = self.run_crash_command(f"ps -c {pid}")
        self.assertIn("(no children)", cmd.stdout)

    def test_times(self):
        cmd = self.check_crash_command(f"ps -t {os.getpid()}")
        foreach_cmd = self.check_crash_command(
            f"foreach {os.getpid()} ps -t", mode="capture"
        )

        for c in (cmd, foreach_cmd):
            self.assertRegex(
                c.stdout,
                r"(?m)^\s*RUN TIME: (?:[0-9] days, )?[0-9]{2}:[0-9]{2}:[0-9]{2}",
            )
            self.assertRegex(c.stdout, r"(?m)^\s*START TIME: [0-9]+$")
            self.assertRegex(c.stdout, r"(?m)^\s*UTIME: [0-9]+$")
            self.assertRegex(c.stdout, r"(?m)^\s*STIME: [0-9]+$")

        for variable in (
            "run_time",
            "start_time",
            "utime",
            "stime",
        ):
            with self.subTest(variable=variable):
                self.assertIsInstance(cmd.drgn_option.globals[variable], Object)
        self._test_drgn_task_header(cmd)

        self.assertEqual(cmd.drgn_option.stdout, foreach_cmd.drgn_option.stdout)

    def test_last_arrival_timestamp(self):
        cmd = self.check_crash_command("ps -l")
        foreach_cmd = self.check_crash_command("foreach ps -l", mode="capture")

        for c in (cmd, foreach_cmd):
            timestamps = []
            for line in c.stdout.splitlines():
                timestamps.append(
                    int(re.match(r"\[\s*([0-9]+)\] \[[A-Z]\]  PID: ", line).group(1))
                )
            self.assertEqual(timestamps, sorted(timestamps, reverse=True))

        self.assertIsInstance(cmd.drgn_option.globals["last_arrival"], Object)
        self.assertIsInstance(cmd.drgn_option.globals["state"], str)
        self._test_drgn_task_header(cmd)

        self.assertEqual(cmd.drgn_option.stdout, foreach_cmd.drgn_option.stdout)

    def test_last_arrival_elapsed(self):
        cmd = self.check_crash_command("ps -m")
        foreach_cmd = self.check_crash_command("foreach ps -m", mode="capture")

        for c in (cmd, foreach_cmd):
            for line in c.stdout.splitlines():
                self.assertRegex(
                    line,
                    r"^\[\s*[0-9]+ [0-9]{2}:[0-9]{2}:[0-9]{2}\.[0-9]{3}\] \[[A-Z]\]  PID: ",
                )

        self.assertIn("task_since_last_arrival_ns(", cmd.drgn_option.stdout)
        self.assertIsInstance(cmd.drgn_option.globals["elapsed"], int)
        self.assertIsInstance(cmd.drgn_option.globals["state"], str)
        self._test_drgn_task_header(cmd)

        self.assertEqual(cmd.drgn_option.stdout, foreach_cmd.drgn_option.stdout)

    def test_last_arrival_cpus(self):
        cpu = min(online_cpus())
        cmd = self.run_crash_command(f"ps -l -C {cpu}")
        foreach_cmd = self.run_crash_command(f"foreach ps -l -C {cpu}")

        for c in (cmd, foreach_cmd):
            timestamps = []
            lines = c.stdout.splitlines()
            self.assertEqual(lines[0], f"CPU: {cpu}")
            for line in lines[1:]:
                timestamps.append(
                    int(
                        re.match(
                            rf"\[\s*([0-9]+)\] \[[A-Z]\]  PID: .*\bCPU: {cpu}\b", line
                        ).group(1)
                    )
                )
            self.assertEqual(timestamps, sorted(timestamps, reverse=True))

    @skip_unless_have_full_mm_support
    @skip_if_highmem
    def test_arguments(self):
        cmd = self.check_crash_command(f"ps -a {os.getpid()}")
        foreach_cmd = self.check_crash_command(
            f"foreach {os.getpid()} ps -a", mode="capture"
        )

        cmdline = Path("/proc/self/cmdline").read_text()
        if cmdline.endswith("\0"):
            cmdline = cmdline[:-1]
        cmdline = cmdline.replace("\0", " ")

        environ = Path("/proc/self/environ").read_text()
        if environ.endswith("\0"):
            environ = environ[:-1]
        environ = environ.replace("\0", "\n     ")

        for c in (cmd, foreach_cmd):
            self.assertRegex(c.stdout, rf"(?m)^ARG: {re.escape(cmdline)}$")
            self.assertRegex(c.stdout, rf"(?m)^ENV: {re.escape(environ)}$")

        self.assertIn("cmdline(", cmd.drgn_option.stdout)
        self.assertIn("environ(", cmd.drgn_option.stdout)
        self.assertIsInstance(cmd.drgn_option.globals["arg"], list)
        self.assertIsInstance(cmd.drgn_option.globals["env"], list)
        self._test_drgn_task_header(cmd)

        self.assertEqual(cmd.drgn_option.stdout, foreach_cmd.drgn_option.stdout)

    def test_thread_groups(self):
        cmd = self.check_crash_command("ps -g")
        foreach_cmd = self.check_crash_command("foreach ps -g", mode="capture")

        for c in (cmd, foreach_cmd):
            self.assertRegex(
                c.stdout,
                rf"(?m)^PID: {os.getpid()}\b.*\n(?:  PID: .*\n)*  PID: {self.threads[0].native_id}\b",
            )
            self.assertNotRegex(c.stdout, rf"(?m)^PID: {self.threads[0].native_id}\b")

        self.assertIn("if not thread_group_leader(task):", cmd.drgn_option.stdout)
        self.assertIn("continue", cmd.drgn_option.stdout)
        self.assertIn("for_each_task_in_group(", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["thread"].type_.type_name(), "struct task_struct *"
        )
        self._test_drgn_task_header(cmd)

        self.assertEqual(cmd.drgn_option.stdout, foreach_cmd.drgn_option.stdout)

    def test_thread_groups_and_pid(self):
        cmd = self.check_crash_command(f"ps -g {os.getpid()}")

        self.assertRegex(
            cmd.stdout,
            rf"(?m)^PID: {os.getpid()}\b.*\n(?:  PID: .*\n)*  PID: {self.threads[0].native_id}\b",
        )
        self.assertNotRegex(cmd.stdout, rf"(?m)^PID: {self.threads[0].native_id}\b")

        self.assertIn("if not thread_group_leader(task):", cmd.drgn_option.stdout)
        self.assertIn("task = task.group_leader", cmd.drgn_option.stdout)
        self.assertIn("for_each_task_in_group(", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["thread"].type_.type_name(), "struct task_struct *"
        )
        self._test_drgn_task_header(cmd)

    def test_thread_groups_no_threads(self):
        with fork_and_stop() as pid:
            cmd = self.run_crash_command(f"ps -g {pid}")

            self.assertIn("(no threads)", cmd.stdout)

    def test_thread_groups_and_tid(self):
        cmd = self.check_crash_command(f"ps -g {self.threads[-1].native_id}")

        self.assertRegex(
            cmd.stdout,
            rf"(?m)^PID: {os.getpid()}\b.*\n(?:  PID: .*\n)*  PID: {self.threads[0].native_id}\b",
        )
        self.assertNotRegex(cmd.stdout, rf"(?m)^PID: {self.threads[-1].native_id}\b")

        self.assertIn("if not thread_group_leader(task):", cmd.drgn_option.stdout)
        self.assertIn("task = task.group_leader", cmd.drgn_option.stdout)
        self.assertIn("for_each_task_in_group(", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["thread"].type_.type_name(), "struct task_struct *"
        )
        self._test_drgn_task_header(cmd)

    def test_thread_groups_and_command_name(self):
        cmd = self.check_crash_command(f"ps -g {_quoted_comm()}")

        self.assertRegex(
            cmd.stdout,
            rf"(?m)^PID: {os.getpid()}\b.*\n(?:  PID: .*\n)*  PID: {self.threads[0].native_id}\b",
        )
        self.assertNotRegex(cmd.stdout, rf"(?m)^PID: {self.threads[0].native_id}\b")

        self.assertIn("if not thread_group_leader(task):", cmd.drgn_option.stdout)
        self.assertIn("continue", cmd.drgn_option.stdout)
        self.assertNotIn("task = task.group_leader", cmd.drgn_option.stdout)
        self.assertIn("for_each_task_in_group(", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["thread"].type_.type_name(), "struct task_struct *"
        )
        self._test_drgn_task_header(cmd)

    def test_thread_groups_and_thread_command_name(self):
        cmd = self.check_crash_command(
            f"ps -g {_quoted_comm(self.threads[0].native_id)}"
        )

        self.assertNotRegex(cmd.stdout, rf"(?m)^PID: {os.getpid()}\b")
        self.assertNotRegex(cmd.stdout, rf"(?m)^PID: {self.threads[0].native_id}\b")

        self.assertIn("if not thread_group_leader(task):", cmd.drgn_option.stdout)
        self.assertIn("continue", cmd.drgn_option.stdout)
        self.assertNotIn("task = task.group_leader", cmd.drgn_option.stdout)
        self.assertIn("for_each_task_in_group(", cmd.drgn_option.stdout)
        self.assertNotIn("thread", cmd.drgn_option.globals)

    def test_rlimit(self):
        cmd = self.check_crash_command(f"ps -r {os.getpid()}")
        foreach_cmd = self.check_crash_command(
            f"foreach {os.getpid()} ps -r", mode="capture"
        )

        for c in (cmd, foreach_cmd):
            for name in (
                "CPU",
                "FSIZE",
                "DATA",
                "STACK",
                "CORE",
                "RSS",
                "NPROC",
                "NOFILE",
                "MEMLOCK",
                "AS",
                "LOCKS",
                "SIGPENDING",
                "MSGQUEUE",
                "NICE",
                "RTPRIO",
                "RTTIME",
            ):
                self.assertRegex(
                    c.stdout, rf"(?m)^\s*{name}(?:\s+(?:[0-9]+|\(unlimited\))){{2}}$"
                )

        self.assertIn("task_rlimits(", cmd.drgn_option.stdout)
        self.assertIsInstance(cmd.drgn_option.globals["rlimits"], dict)
        self._test_drgn_task_header(cmd)

        self.assertEqual(cmd.drgn_option.stdout, foreach_cmd.drgn_option.stdout)

    def test_summary(self):
        cmd = self.check_crash_command("ps -S")
        foreach_cmd = self.check_crash_command("foreach ps -S", mode="capture")

        for c in (cmd, foreach_cmd):
            self.assertRegex(c.stdout, r"^(?:\s*[A-Z]: [0-9]+)+$")

        self.assertIn("task_state_to_char(", cmd.drgn_option.stdout)
        self.assertIsInstance(cmd.drgn_option.globals["counter"], collections.Counter)

        self.assertEqual(cmd.drgn_option.stdout, foreach_cmd.drgn_option.stdout)


class TestForeach(CrashCommandTestCase):
    def test_help(self):
        # The actual functionality of foreach is tested in the individual
        # subcommands. This just tests that we properly document what
        # subcommands are supported.
        cmd = self.run_crash_command("help foreach")
        match = re.search(
            r"Currently,\s+((?:\*\*\w+\*\*,\s+)+)and\s+(\*\*\w+\*\*)\s+are\s+supported",
            cmd.stdout,
        )
        self.assertTrue(match)
        documented = (
            (match.group(1) + match.group(2)).replace("*", "").replace(",", "").split()
        )
        self.assertEqual(documented, sorted(_CRASH_FOREACH_SUBCOMMANDS))



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import contextlib
import errno
import os
from unittest import SkipTest

from tests.linux_kernel.bpf import (
    BPF_LD_MAP_FD,
    BPF_MAP_TYPE_HASH,
    BPF_PROG_TYPE_KPROBE,
    BPF_PROG_TYPE_SOCKET_FILTER,
    BPF_REG_0,
    bpf_map_create,
    bpf_map_get_info_by_fd,
    bpf_prog_get_info_by_fd,
    bpf_prog_ids,
    bpf_prog_load,
)
from tests.linux_kernel.crash_commands import CrashCommandTestCase
from tests.linux_kernel.helpers.test_bpf import BpfTestCase


class TestBpf(CrashCommandTestCase, BpfTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        # This command isn't supported before Linux kernel commits dc4bb0e23561
        # ("bpf: Introduce bpf_prog ID") and f3f1c054c288 ("bpf: Introduce bpf_map
        # ID") (in v4.13).
        try:
            next(bpf_prog_ids())
        except StopIteration:
            # Kernel supports BPF IDs but no programs exist yet
            pass
        except OSError as e:
            if e.errno != errno.EINVAL:
                raise
            raise SkipTest("This kernel version doesn't support BPF object IDs")

    def test_bpf(self):
        with contextlib.ExitStack() as exit_stack:
            map_fd = bpf_map_create(BPF_MAP_TYPE_HASH, 8, 8, 8)
            exit_stack.callback(os.close, map_fd)

            prog1_fd = bpf_prog_load(BPF_PROG_TYPE_KPROBE, self.INSNS, b"GPL")
            exit_stack.callback(os.close, prog1_fd)

            prog2_insns = BPF_LD_MAP_FD(BPF_REG_0, map_fd) + self.INSNS
            prog2_fd = bpf_prog_load(BPF_PROG_TYPE_SOCKET_FILTER, prog2_insns, b"GPL")
            exit_stack.callback(os.close, prog2_fd)

            map_id = bpf_map_get_info_by_fd(map_fd).id
            prog1_id = bpf_prog_get_info_by_fd(prog1_fd).id
            prog2_info = bpf_prog_get_info_by_fd(prog2_fd)
            prog2_id = prog2_info.id
            prog2_tag = "".join(f"{b:02x}" for b in prog2_info.tag)

            cmd = self.check_crash_command("bpf")

            self.assertIn("BPF_PROG_TYPE", cmd.stdout)
            self.assertRegex(
                cmd.stdout,
                rf"(?sm)BPF_PROG.*^\s*{prog1_id}\s+.*KPROBE\s+[0-9a-f]{{16}}\s*$",
            )
            self.assertRegex(
                cmd.stdout,
                rf"(?sm)BPF_PROG.*^\s*{prog2_id}\s+.*SOCKET_FILTER\s+.*{prog2_tag}\s+{map_id}$",
            )
            self.assertIn("BPF_MAP_TYPE", cmd.stdout)
            self.assertRegex(
                cmd.stdout,
                rf"(?sm)BPF_MAP.*^\s*{map_id}\s+.*HASH\s+",
            )

    def test_bpf_program_by_id(self):
        with contextlib.ExitStack() as exit_stack:
            prog_fd = bpf_prog_load(BPF_PROG_TYPE_KPROBE, self.INSNS, b"GPL")
            exit_stack.callback(os.close, prog_fd)

            prog_info = bpf_prog_get_info_by_fd(prog_fd)
            prog_id = prog_info.id
            prog_tag = "".join(f"{b:02x}" for b in prog_info.tag)

            cmd = self.check_crash_command(f"bpf -p {prog_id}")

            self.assertRegex(
                cmd.stdout,
                rf"(?sm)ID\s+.*BPF_PROG_TYPE.*^\s*{prog_id}\s+.*KPROBE\s+.*{prog_tag}\s*$",
            )

            self.assertRegex(
                cmd.stdout,
                r"(?sm)^\s*XLATED:\s*\d+\s+JITED:\s*\d+\s+MEMLOCK:\s*\d+",
            )

            self.assertRegex(
                cmd.stdout,
                r"(?sm)^\s*LOAD_TIME:\s*((\w+\s+){3}\d{2}:\d{2}:\d{2}\s+\d{4}|\(unknown\))",
            )

            self.assertRegex(
                cmd.stdout,
                r"(?sm)^\s*GPL_COMPATIBLE:\s*(yes|no)\s+NAME:\s*(\(unused\)|\(unknown\)|\S+)\s+UID:\s*\d+",
            )

    def test_bpf_program_by_id_invalid(self):
        self.assertRaises(
            Exception,
            self.check_crash_command,
            "bpf -p 1a",
        )
        self.assertRaises(
            Exception,
            self.check_crash_command,
            "bpf -p 1$^7",
        )

    def test_bpf_map_by_id(self):
        with contextlib.ExitStack() as exit_stack:
            map_fd = bpf_map_create(BPF_MAP_TYPE_HASH, 4, 8, 256)
            exit_stack.callback(os.close, map_fd)

            map_info = bpf_map_get_info_by_fd(map_fd)
            map_id = map_info.id

            cmd = self.check_crash_command(f"bpf -m {map_id}")

            self.assertIn("BPF_MAP_TYPE", cmd.stdout)

            self.assertRegex(
                cmd.stdout,
                rf"(?sm)ID\s+.*BPF_MAP.*^\s*{map_id}\s+.*HASH\s+",
            )

            self.assertRegex(
                cmd.stdout,
                r"(?sm)^\s*KEY_SIZE:\s*4\s+VALUE_SIZE:\s*8\s+MAX_ENTRIES:\s*256",
            )

            self.assertRegex(
                cmd.stdout,
                r"(?sm)NAME:\s*(\(unused\)|\(unknown\)|\"[^\"]*\")\s+UID:\s*(\d+|\(unknown\)|\(unused\))",
            )

    def test_bpf_map_by_id_with_program(self):
        with contextlib.ExitStack() as exit_stack:
            map_fd = bpf_map_create(BPF_MAP_TYPE_HASH, 8, 8, 16)
            exit_stack.callback(os.close, map_fd)

            prog_insns = BPF_LD_MAP_FD(BPF_REG_0, map_fd) + self.INSNS
            prog_fd = bpf_prog_load(BPF_PROG_TYPE_SOCKET_FILTER, prog_insns, b"GPL")
            exit_stack.callback(os.close, prog_fd)

            map_info = bpf_map_get_info_by_fd(map_fd)
            map_id = map_info.id

            cmd = self.check_crash_command(f"bpf -m {map_id}")

            self.assertIn(f"{map_id}", cmd.stdout)
            self.assertIn("HASH", cmd.stdout)
            self.assertIn("KEY_SIZE:", cmd.stdout)
            self.assertIn("VALUE_SIZE:", cmd.stdout)
            self.assertIn("MAX_ENTRIES:", cmd.stdout)

    def test_bpf_map_by_id_invalid(self):
        self.assertRaises(
            Exception,
            self.check_crash_command,
            "bpf -m abc",
        )

        self.assertRaises(
            Exception,
            self.check_crash_command,
            "bpf -m 1$^7",
        )

    def test_bpf_command_shows_maps(self):
        with contextlib.ExitStack() as exit_stack:
            map_fd = bpf_map_create(BPF_MAP_TYPE_HASH, 8, 8, 8)
            exit_stack.callback(os.close, map_fd)

            map_id = bpf_map_get_info_by_fd(map_fd).id

            cmd = self.check_crash_command("bpf")

            self.assertIn("BPF_MAP", cmd.stdout)

            self.assertRegex(
                cmd.stdout,
                rf"(?sm)^\s*{map_id}\s+.*HASH\s+",
            )



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import signal
import unittest

import drgn
from drgn import Object, sizeof
from drgn.helpers.linux.mm import follow_phys
from tests.linux_kernel import (
    skip_unless_have_full_mm_support,
    skip_unless_have_test_kmod,
)
from tests.linux_kernel.crash_commands import CrashCommandTestCase


@skip_unless_have_full_mm_support
@skip_unless_have_test_kmod
class TestSearch(CrashCommandTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.subject = cls.prog["drgn_test_search_subject"]
        cls.subject_args = f"-s {cls.subject.address_:x} -l {sizeof(cls.subject)}"

    def test_long(self):
        cmd = self.check_crash_command(f"search {self.subject_args} deadbeef")
        self.assertRegex(
            cmd.stdout, rf"(?m)^\s*{self.subject.l.address_:x}:\s+deadbeef$"
        )

        self.assertIn("search_memory_word(0xdeadbeef)", cmd.drgn_option.stdout)
        self.assertEqual(cmd.drgn_option.globals["address"], self.subject.l.address_)
        self.assertEqual(cmd.drgn_option.globals["value"], 0xDEADBEEF)

    def test_int(self):
        cmd = self.check_crash_command(f"search {self.subject_args} -w deadb00")
        self.assertRegex(
            cmd.stdout, rf"(?m)^\s*{self.subject.i.address_:x}:\s+deadb00$"
        )

        self.assertIn("search_memory_u32(0xdeadb00)", cmd.drgn_option.stdout)
        self.assertEqual(cmd.drgn_option.globals["address"], self.subject.i.address_)
        self.assertEqual(cmd.drgn_option.globals["value"], 0xDEADB00)

    def test_short(self):
        cmd = self.check_crash_command(f"search {self.subject_args} -h b0ba")
        self.assertRegex(cmd.stdout, rf"(?m)^\s*{self.subject.s.address_:x}:\s+b0ba$")

        self.assertIn("search_memory_u16(0xb0ba)", cmd.drgn_option.stdout)
        self.assertEqual(cmd.drgn_option.globals["address"], self.subject.s.address_)
        self.assertEqual(cmd.drgn_option.globals["value"], 0xB0BA)

    def test_multiple(self):
        cmd = self.check_crash_command(f"search {self.subject_args} deadbeef fee1dead")
        self.assertRegex(
            cmd.stdout, rf"(?m)^\s*{self.subject.l.address_:x}:\s+deadbeef$"
        )
        self.assertRegex(
            cmd.stdout, rf"(?m)^\s*{self.subject.l2.address_:x}:\s+fee1dead$"
        )

        self.assertIn(
            "search_memory_word(0xdeadbeef, 0xfee1dead)", cmd.drgn_option.stdout
        )
        self.assertIn(
            cmd.drgn_option.globals["address"],
            (self.subject.l.address_, self.subject.l2.address_),
        )
        self.assertIn(cmd.drgn_option.globals["value"], (0xDEADBEEF, 0xFEE1DEAD))

    @unittest.skipUnless(drgn._with_pcre2, "built without pcre2 support")
    def test_string(self):
        cmd = self.check_crash_command(f"search {self.subject_args} -c 'hello, world!'")
        self.assertRegex(
            cmd.stdout,
            rf"(?m)^\s*{self.subject.c.address_:x}:\s+hello, world!\.{{43}}$",
        )

        self.assertIn('search_memory("hello, world!")', cmd.drgn_option.stdout)
        self.assertEqual(cmd.drgn_option.globals["address"], self.subject.c.address_)

    @unittest.skipUnless(drgn._with_pcre2, "built without pcre2 support")
    def test_multiple_strings(self):
        cmd = self.check_crash_command(
            f"search {self.subject_args} -c 'hello, world!' 'goodbye, world!'"
        )
        self.assertRegex(
            cmd.stdout,
            rf"(?m)^\s*{self.subject.c.address_:x}:\s+hello, world!\.{{43}}$",
        )
        self.assertRegex(
            cmd.stdout,
            rf"(?m)^\s*{self.subject.c2.address_:x}:\s+goodbye, world!\.{{41}}$",
        )

        self.assertIn(
            'search_memory_regex(rb"hello,\\ world!|goodbye,\\ world!")',
            cmd.drgn_option.stdout,
        )
        self.assertIn(
            cmd.drgn_option.globals["address"],
            (self.subject.c.address_, self.subject.c2.address_),
        )
        self.assertIn(
            cmd.drgn_option.globals["value"], (b"hello, world!", b"goodbye, world!")
        )

    def test_context(self):
        cmd = self.check_crash_command(f"search {self.subject_args} fee1dead -x 1")
        self.assertRegex(
            cmd.stdout, rf"(?m)^\s*{self.subject.l2.address_:x}:\s+fee1dead$"
        )
        self.assertIn("deadbeef", cmd.stdout)
        self.assertIn("deadb00", cmd.stdout)

        self.assertIn("search_memory_word(0xfee1dead)", cmd.drgn_option.stdout)
        self.assertEqual(cmd.drgn_option.globals["address"], self.subject.l2.address_)
        self.assertEqual(cmd.drgn_option.globals["value"], 0xFEE1DEAD)
        # Note: we don't do anything in the --drgn output for -x.

    def test_ignore_mask(self):
        cmd = self.check_crash_command(
            f"search {self.subject_args} 0e000e00 -m f0fff0ff"
        )
        self.assertRegex(
            cmd.stdout, rf"(?m)^\s*{self.subject.l.address_:x}:\s+deadbeef$"
        )
        self.assertRegex(
            cmd.stdout, rf"(?m)^\s*{self.subject.l2.address_:x}:\s+fee1dead$"
        )

        self.assertIn(
            "search_memory_word(0xe000e00, ignore_mask=0xf0fff0ff)",
            cmd.drgn_option.stdout,
        )
        self.assertIn(
            cmd.drgn_option.globals["address"],
            (self.subject.l.address_, self.subject.l2.address_),
        )
        self.assertIn(cmd.drgn_option.globals["value"], (0xDEADBEEF, 0xFEE1DEAD))

    def test_symbols(self):
        cmd = self.check_crash_command(
            f"search -s drgn_test_search_subject -l {sizeof(self.subject)} drgn_test_function"
        )
        sym_address = self.prog.symbol("drgn_test_function").address
        self.assertRegex(
            cmd.stdout,
            rf"(?m)^\s*{self.subject.func.address_:x}:\s+{sym_address:x}\s+\(drgn_test_function\)$",
        )

        self.assertIn(
            'search_memory_word(prog.symbol("drgn_test_function").address)',
            cmd.drgn_option.stdout,
        )
        self.assertIn(
            'prog.symbol("drgn_test_search_subject").address', cmd.drgn_option.stdout
        )
        self.assertEqual(cmd.drgn_option.globals["address"], self.subject.func.address_)
        self.assertEqual(cmd.drgn_option.globals["value"], sym_address)

    def test_end(self):
        start = self.subject.address_
        end = start + sizeof(self.subject)
        cmd = self.check_crash_command(f"search -s {start:x} -e {end:x} deadbeef")
        self.assertRegex(
            cmd.stdout, rf"(?m)^\s*{self.subject.l.address_:x}:\s+deadbeef$"
        )

        self.assertIn("search_memory_word(0xdeadbeef)", cmd.drgn_option.stdout)
        self.assertIn(
            f"set_address_range(min_address={start:#x}, max_address={end:#x} - 1)",
            cmd.drgn_option.stdout,
        )
        self.assertEqual(cmd.drgn_option.globals["address"], self.subject.l.address_)
        self.assertEqual(cmd.drgn_option.globals["value"], 0xDEADBEEF)

    def test_physical(self):
        subject = self.prog["drgn_test_search_subject2"]
        address = follow_phys(self.prog["init_mm"].address_of_(), subject).value_()
        cmd = self.check_crash_command(
            f"search -p -s {address:x} -l {sizeof(subject[0])} 12345678"
        )
        self.assertRegex(cmd.stdout, rf"(?m)^\s*{address:x}:\s+12345678$")

        self.assertIn("search_memory_word(0x12345678)", cmd.drgn_option.stdout)
        self.assertIn("physical=True", cmd.drgn_option.stdout)
        self.assertEqual(cmd.drgn_option.globals["address"], address)
        self.assertEqual(cmd.drgn_option.globals["value"], 0x12345678)

    def test_task_stacks(self):
        value = self.prog["drgn_test_small_slab_objects"][0].value_()
        cmd = self.check_crash_command(f"search -t {value:x}")
        # drgn_test_kthread gets truncated to 16 bytes (including the null
        # terminator).
        self.assertIn("drgn_test_kthre", cmd.stdout)

        self.assertIn(f"search_memory_word({value:#x})", cmd.drgn_option.stdout)
        self.assertIn("for_each_task(", cmd.drgn_option.stdout)
        self.assertIn("set_address_range(stack", cmd.drgn_option.stdout)
        self.assertIsInstance(cmd.drgn_option.globals["task"], Object)

    def test_active_task_stacks(self):
        # We don't have a good way to exercise this, so just test that it runs.
        self.check_crash_command("search -T deadbeef")

    def test_no_start_or_end(self):
        def handler(signum, frame):
            raise TimeoutError()

        old_handler = signal.signal(signal.SIGALRM, handler)
        self.addCleanup(signal.signal, signal.SIGALRM, old_handler)

        try:
            # We don't want to wait to search all of memory, so just run the
            # code for a short time.
            signal.setitimer(signal.ITIMER_REAL, 0.1)
            self.run_crash_command_drgn_option("search deadbeef")
        except TimeoutError:
            pass
        finally:
            signal.setitimer(signal.ITIMER_REAL, 0)
